'''save/load from MongoDB routines

-------
'''

# to start a mongodb server on the local workstation
# mongod --dbpath $DIRECTORY_WHERE_TO_STORE_DATA

from __future__ import print_function, division, unicode_literals

from .omas_utils import *
from .omas_core import ODS


# -----------------------------
# save and load OMAS to MongoDB
# -----------------------------
def save_omas_mongo(ods, collection, database='omas', server=omas_rcparams['default_mongo_server']):
    """
    Save an OMAS data set to MongoDB

    :param ods: OMAS data set

    :param collection: collection name in the database

    :param database: database name on the server

    :param server: server name

    :return: unique `_id` identifier of the record
    """

    printd('Saving OMAS data to MongoDB: collection=%s database=%s  server=%s' % (collection, database, server), topic=['MongoDB'])

    # importing module
    from pymongo import MongoClient

    # connect
    client = MongoClient(server.format(**get_mongo_credentials(server, database, collection)))

    # access database
    db = client[database]

    # access collection
    coll = db[collection]

    # a cheap way to encode data
    kw = {'indent': 0, 'separators': (',', ': '), 'sort_keys': True}
    json_string = json.dumps(ods, default=lambda x: json_dumper(x, None), **kw)
    data = json.loads(json_string)

    # avoid insert_one() to modify data
    data = copy.copy(data)

    # insert record
    res = coll.insert_one(data)
    return str(res.inserted_id)


def load_omas_mongo(find, collection, database='omas', server=omas_rcparams['default_mongo_server'], consistency_check=True, imas_version=omas_rcparams['default_imas_version']):
    """
    Load an OMAS data set from MongoDB

    :param find: dictionary to find data in the database

    :param collection: collection name in the database

    :param database: database name on the server

    :param server: server name

    :param consistency_check: verify that data is consistent with IMAS schema

    :param imas_version: imas version to use for consistency check

    :return: list of OMAS data set that match find criterion
    """

    # importing module
    from pymongo import MongoClient
    from bson.objectid import ObjectId

    # allow search by _id
    if not isinstance(find, dict):
        raise TypeError('load_omas_mongo find attribute must be a dictionary')
    if '_id' in find:
        find = copy.deepcopy(find)
        find['_id'] = ObjectId(find['_id'])

    printd('Loading OMAS data from MongoDB: collection=%s database=%s  server=%s' % (collection, database, server), topic=['MongoDB'])

    # connect
    client = MongoClient(server.format(**get_mongo_credentials(server, database, collection)))

    # access database
    db = client[database]

    # access collection
    coll = db[collection]

    # find all the matching records
    results = {}
    for record in coll.find(find):
        ods = ODS(consistency_check=consistency_check, imas_version=imas_version)
        _id = record['_id']
        del record['_id']
        ods.from_structure(record)
        results[_id] = ods

    return results


def get_mongo_credentials(server='', database='', collection=''):
    '''
    Users can specify their credentials in a `~/.omas/mongo_credentials` json file
    formatted like this:

    >> {"default": {"user": "mydefaultuser",
    >>              "pass": "mydefaultpass"},
    >>              "omasdb-xymmt.mongodb.net": {"user": "myuser1"
    >>                                           "pass": "mypass1",
    >>                                           "specific_database": {"user": "myuser2",
    >>                                                                 "pass": "mypass2",
    >>                                                                 "specific_collection": {"user": "myuser3",
    >>                                                                                         "pass": "mypass3"}
    >> }}}

    if no `~/.omas/mongo_credentials` file is found: {'user': 'omas_test', 'pass': 'omas_test'}

    if no matching server is found, the `default` is returned

    :param server: server to look credentials for
        * server can have `{user}:{pass}` placeholders: mongodb+srv://{user}:{pass}@omasdb-xymmt.mongodb.net

    :param database: database name in server to look credentials for

    :param collection: collection name in database to look credentials for

    :return: dictionary with 'user' and 'pass' keys
    '''
    server = server.split('@')[-1]
    up = {'user': 'omas_test', 'pass': 'omas_test'}
    config = {}
    filename = os.environ['HOME'] + '/.omas/mongo_credentials'
    if os.path.exists(filename):
        with open(filename) as f:
            config = json.loads(f.read())
    if 'default' in config:
        up = config['default']
    if server in config:
        up = config[server]
        if database in config[server]:
            up = config[server][database]
            if collection in config[server][collection]:
                up = config[server][database][collection]
    return up


def through_omas_mongo(ods, method=['function', 'class_method'][1]):
    """
    Test save and load OMAS MongoDB

    :param ods: ods

    :return: ods
    """
    if method == 'function':
        _id = save_omas_mongo(ods, collection='test', database='test')
        results = load_omas_mongo({'_id': _id}, collection='test', database='test')
        if len(results) != 1:
            raise Exception('through_omas_mongo failed')
        ods1 = list(results.values())[0]
        return ods1
    else:
        _id = ods.save('mongo', collection='test', database='test')
        ods1 = ODS().load('mongo', {'_id': _id}, collection='test', database='test')
        return ods1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Units conversions
=================
This example illustrates how OMAS can automatically translate units by leveraging the `pint` Python package
"""

from __future__ import print_function, division, unicode_literals

from omas import *

import pint

ureg = pint.UnitRegistry()

ods = ODS()

invalue = 8.0 * ureg.milliseconds
print('Data input: %s' % invalue)
ods['equilibrium.time_slice[0].constraints.diamagnetic_flux.time_measurement'] = invalue

outvalue = ods['equilibrium.time_slice[0].constraints.diamagnetic_flux.time_measurement']
print('Data output without units support: %s (IMAS uses MKS)' % outvalue)

with omas_environment(ods, unitsio=True):
    outvalue = ods['equilibrium.time_slice[0].constraints.diamagnetic_flux.time_measurement']
print('Data output with units support: %s' % outvalue)

###################################
# .. code-block:: none
#
#     Data input: 8.0 millisecond
#     Data output without units support: 0.008 (IMAS uses MKS)
#     Data output with units support: 0.008 second
#



#!/usr/bin/env python
# # -*- coding: utf-8 -*-
"""
MongoDB storage and discovery
=============================
OMAS can store/load its data in a MongoDB server.
MongoDB is a document-oriented database program that uses JSON-like documents with schema
and supports field, range query, and regular expression searches.

As an example we illustrate storage of GKDB data in this format.
"""

from pprint import *
from omas import *

ods = ODS().sample_equilibrium()
ods['equilibrium.code.name'] = 'test_code'

# write entry to the database
_id = save_omas_mongo(ods, collection='test', database='test')

# retrieve exact entry using `_id`
ods = ODS().load('mongo', {'_id': _id}, collection='test', database='test')

# find entries that satisfy the query by matching strings
odss = load_omas_mongo({'equilibrium.code.name': 'test_code'}, collection='test', database='test')

# navigate through dictionaries and arrays of structures
# https://docs.mongodb.com/manual/tutorial/query-embedded-documents/
odss = load_omas_mongo({'equilibrium.time_slice.0.global_quantities.ip': {'$gt': 0}}, collection='test', database='test')

# find entries based on conditions on array elements
# https://docs.mongodb.com/manual/tutorial/query-arrays/
odss = load_omas_mongo({'equilibrium.vacuum_toroidal_field.b0': {'$size': 1}}, collection='test', database='test')

# =============================================
# showcase use of MongoDB storage for GKDB data
# =============================================
# load a sample GKDB sample json file
# note use of `consistency_check='warn'` is necessary since GKDB json is in fact not yet 100% compatible with IMAS
sample_filename = imas_json_dir + '/../samples/gkdb_linear_eigenvalue.json'
ods = ODS(consistency_check='warn')
ods['gyrokinetics'].load(sample_filename)

# write GKDB entry to the database
_id = ods.save('mongo', collection='gkdb', database='test')

# reload GKDB entry
ods1 = ODS(consistency_check='warn')
ods1.load('mongo', {'_id': _id}, collection='gkdb', database='test')

# look for differences between original GKDB json and MongoDB entry
differences = ods.diff(ods1, ignore_type=True)
if not differences:
    print('\nPrint no differences found: save/load of GKDB on MongoDB worked\n')
else:
    pprint(differences)
    raise RuntimeError('Save/Load of GKDB  on MongoDB failed')



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
OMFIT gEQDSK and IMAS
=====================
This example loads a gEQDSK file using the OMFITgeqdsk class.
The gEQKDS file is then save to IMAS and loaded back.

This example only works with python 2.7 since OMFIT is only Python 2.7 compatible

Prior running this script, the following commands must be typed at the teriminal
> import IMAS OMAS
> imasdb ITER

.. figure:: ../images/eq_omas_omfit.png
  :align: center
  :width: 75%
  :alt: OMFIT+OMAS facilitate save/load gEQDSK to/from IMAS
  :target: /.._images/eq_omas_omfit.png

"""

from __future__ import print_function, division, unicode_literals

import os
from matplotlib import pyplot

from omfit.classes.omfit_eqdsk import OMFITgeqdsk, OMFITsrc
from omas import *

imas_version = os.environ.get('IMAS_VERSION', omas_rcparams['default_imas_version'])

# settings
os.environ['OMAS_DEBUG_TOPIC'] = 'imas'
omas_rcparams['allow_fake_imas_fallback'] = True

# read gEQDSK file in OMFIT
eq = OMFITgeqdsk(OMFITsrc + '/../samples/g133221.01000')

# convert gEQDSK to OMAS data structure
ods = eq.to_omas()

# save OMAS data structure to IMAS
paths = save_omas_imas(ods, machine='DIII-D', pulse=133221, new=True)

# load OMAS data structure from IMAS
ods1 = load_omas_imas(machine='DIII-D', pulse=133221)

# generate gEQDSK file from OMAS data structure
eq1 = OMFITgeqdsk('g133221.02000').from_omas(ods1)

# save gEQDSK file
eq1.deploy('g133221.02000')

# plot
eq.plot()
eq1.plot()
pyplot.show()



'''pypi setup file

-------
'''

from __future__ import print_function, division, unicode_literals

# --------------------------------------------
# external imports
# --------------------------------------------
import os
import sys
import glob
import json
import copy
from collections import MutableMapping, OrderedDict
import re
import numpy
from pprint import pprint
from io import StringIO
from contextlib import contextmanager
import tempfile
import warnings

formatwarning_orig = warnings.formatwarning
warnings.formatwarning = lambda message, category, filename, lineno, line=None: \
    formatwarning_orig(message, category, filename, lineno, line='')

# pint: avoid loading pint upfront since it can be slow and it is not always used
ureg = []
if False:
    import pint

    ureg.append(pint.UnitRegistry())
else:
    ureg.append(None)

# uncertainties
import uncertainties
import uncertainties.unumpy as unumpy
from uncertainties.unumpy import nominal_values, std_devs, uarray
from uncertainties import ufloat
import ast

# xarrays: avoid loading xarrays upfront since it can be slow and it is not always used
import xarray

# Python3/2 import differences
if sys.version_info < (3, 0):
    import cPickle as pickle


    def b2s(string):
        return string

else:
    basestring = str
    unicode = str
    import pickle

    _orig_pickle_loads = pickle.loads


    def _pickle_loads_python2compatible(*args, **kw):
        kw.setdefault('encoding', 'latin1')
        return _orig_pickle_loads(*args, **kw)


    pickle.loads = _pickle_loads_python2compatible

    _orig_pickle_load = pickle.load


    def _pickle_load_python2compatible(*args, **kw):
        kw.setdefault('encoding', 'latin1')
        return _orig_pickle_load(*args, **kw)


    pickle.load = _pickle_load_python2compatible


    def b2s(bytes):
        return bytes.decode("utf-8")


# --------------------------------------------
# configuration of directories and IMAS infos
# --------------------------------------------
class IMAS_json_dir(unicode):
    '''
    directory where the JSON data structures for the different versions of IMAS are stored
    '''
    pass


imas_json_dir = IMAS_json_dir(os.path.abspath(str(os.path.dirname(__file__)) + '/imas_structures/'))


class IMAS_versions(OrderedDict):
    '''
    dictionary with list of IMAS version and their sub-folder name in the imas_json_dir
    '''

    def __init__(self, mode='all'):
        OrderedDict.__init__(self)
        if mode in ['all', 'named']:
            # first `develop/3` and other branches
            for _item in list(map(lambda x: os.path.basename(x), sorted(glob.glob(imas_json_dir + os.sep + '*')))):
                if not _item.startswith('3'):
                    self[_item.replace('_', '.')] = _item
        if mode in ['all', 'tagged']:
            # next all tagged versions sorted by version number
            for _item in list(map(lambda x: os.path.basename(x), sorted(glob.glob(imas_json_dir + os.sep + '*')))):
                if _item.startswith('3'):
                    self[_item.replace('_', '.')] = _item


imas_versions = IMAS_versions()

if 'OMAS_IMAS_VERSION' in os.environ:
    _default_imas_version = os.environ['OMAS_IMAS_VERSION']
else:
    try:
        _default_imas_version = list(imas_versions.keys())[-1]
    except IndexError:
        # IndexError will occur if `imas_json_dir` is empty: we must allow going forward, at least to build_json_structures
        _default_imas_version = ''


# --------------------------------------------
# rcparams
# --------------------------------------------
class OMAS_rc_params(dict):
    '''
    dictionary of parameters that control how OMAS operates
    '''
    pass


omas_rcparams = OMAS_rc_params()
omas_rcparams.update({
    'cocos': 11,
    'cocosio': 11,
    'coordsio': {},
    'consistency_check': True,
    'dynamic_path_creation': True,
    'unitsio': False,
    'tmp_imas_dir': os.environ.get('OMAS_TMP_DIR', os.sep.join([tempfile.gettempdir(),
                                                                os.environ.get('USER', 'dummy_user'),
                                                                'OMAS_TMP_DIR'])),
    'fake_imas_dir': os.environ.get('OMAS_FAKE_IMAS_DIR', os.sep.join([os.environ.get('HOME', tempfile.gettempdir()),
                                                                       'tmp',
                                                                       'OMAS_FAKE_IMAS_DIR'])),
    'allow_fake_imas_fallback': bool(int(os.environ.get('OMAS_ALLOW_FAKE_IMAS_FALLBACK', '0'))),
    'default_imas_version': _default_imas_version,
    'default_mongo_server': 'mongodb+srv://{user}:{pass}@omasdb-xymmt.mongodb.net',
    'pickle_protocol': 2  # pickle.HIGHEST_PROTOCOL (`2` is used to ensure Python 3-->2 compatibility)
})


@contextmanager
def rcparams_environment(**kw):
    old_omas_rcparams = omas_rcparams.copy()
    omas_rcparams.update(kw)
    try:
        yield omas_rcparams
    finally:
        omas_rcparams.update(old_omas_rcparams)


# --------------------------------------------
# additional data structures
# --------------------------------------------
add_datastructures = {}
# info data structure carries shot/run/version/machine/user info through different save formats
add_datastructures['info'] = {
    "info.shot": {
        "full_path": "info.shot",
        "data_type": "INT_0D",
        "description": "shot number"
    },
    "info.imas_version": {
        "full_path": "info.imas_version",
        "data_type": "STR_0D",
        "description": "imas version"
    },
    "info.machine": {
        "full_path": "info.machine",
        "data_type": "STR_0D",
        "description": "machine name"
    },
    "info.user": {
        "full_path": "info.user",
        "data_type": "STR_0D",
        "description": "user name"
    },
    "info.run": {
        "full_path": "info.run",
        "data_type": "INT_0D",
        "description": "run number"
    }
}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
OMAS plot examples
==================
This example loads some data from S3, augments the ODS with pressure information, and generates some plots
"""

from __future__ import print_function, division, unicode_literals

from matplotlib.pyplot import show
from omas import *

# load some data from S3
ods = load_omas_s3('OMFITprofiles_sample', user='omas_shared')

# augment ODS with pressure information
ods.consistency_check = False  # NOTE: some extra pressures are calculated if IMAS consistency_check = False
ods.physics_core_profiles_pressures()

# omas plot for pressures
ods.plot_core_profiles_pressures()
show()

# omas plot for core profiles
ods.plot_core_profiles_summary()
show()

# omas plot for equilibrium
omas_plot.equilibrium_summary(ods, linewidth=1, label='my equilibrium')
show()



#!/usr/bin/env python
# # -*- coding: utf-8 -*-

"""
Test script for omas/omas_utils.py

.. code-block:: none

   python -m unittest omas/tests/test_omas_utils

-------
"""

# Basic imports
from __future__ import print_function, division, unicode_literals
import os
import unittest
import numpy
import warnings
import copy

# Plot imports
import matplotlib as mpl
from matplotlib import pyplot as plt

# OMAS imports
from omas import *
from omas.omas_utils import *


class TestOmasUtils(unittest.TestCase):
    """
    Test suite for omas_utils.py
    """

    # Flags to edit while testing
    verbose = False  # Spammy, but occasionally useful for debugging a weird problem

    # Sample data for use in tests
    specific_test_version = '3.18.0'

    # Utilities for this test
    def printv(self, *arg):
        """Utility for tests to use"""
        if self.verbose:
            print(*arg)

    def setUp(self):
        test_id = self.id()
        test_name = '.'.join(test_id.split('.')[-2:])
        self.printv('{}...'.format(test_name))

    def tearDown(self):
        test_name = '.'.join(self.id().split('.')[-2:])
        self.printv('    {} done.'.format(test_name))

    def test_different_ods(self):
        ods = ODS()
        ods2 = ODS()
        ods2.sample_equilibrium()
        diff_eq = different_ods(ods, ods2)
        self.printv('  diff_eq = {}'.format(diff_eq))
        assert isinstance(diff_eq, list)
        assert ('equilibrium' in ' '.join(diff_eq)) or ('wall' in ' '.join(diff_eq))
        ods3 = copy.deepcopy(ods2)
        assert different_ods(ods2, ods3) is False
        ods3.sample_core_profiles()
        diff_prof = ods2.diff(ods3)
        self.printv('  diff_prof = {}'.format(diff_prof))
        assert isinstance(diff_prof, list)
        assert isinstance(different_ods(ods3, ods2), list)
        assert 'core_profiles' in ' '.join(diff_prof)
        ods2.sample_core_profiles(include_pressure=False)
        diff_prof2 = different_ods(ods3, ods2)
        self.printv('  diff_prof2 = {}'.format(diff_prof2))
        assert isinstance(diff_prof2, list)
        assert 'core_profiles' in ' '.join(diff_prof2)
        ods2.sample_core_profiles()
        ods2['core_profiles.profiles_1d.0.electrons.density_thermal'][0] = 1.5212
        diff_prof3 = ods2.diff(ods3)
        self.printv('  diff_prof3 = {}'.format(diff_prof3))
        assert isinstance(diff_prof3, list)
        assert 'value' in ' '.join(diff_prof3)
        ods2.sample_core_profiles()
        ods2['core_profiles.profiles_1d.0.ion.0.element.0.a'] = 2
        diff_prof4 = different_ods(ods2, ods3)
        assert not diff_prof4
        ods2.sample_core_profiles()
        ods2['core_profiles.code.name'] = 'fake name 1'
        ods3['core_profiles.code.name'] = 'fake name 2'
        diff_prof5 = different_ods(ods2, ods3)
        self.printv('  diff_prof5 = {}'.format(diff_prof5))
        assert isinstance(diff_prof5, list)
        assert 'name' in ' '.join(diff_prof5)

    def test_printe(self):
        printe('printe_test,', end='')
        printw('printw_test', end='')

    def test_is_numeric(self):
        assert is_numeric(5) is True
        assert is_numeric(numpy.array([5])) is True
        assert is_numeric('blah') is False
        assert is_numeric({'blah': 'blah'}) is False
        assert is_numeric([]) is False
        assert is_numeric(None) is False

    def test_remove_parentheses(self):
        assert remove_parentheses('zoom(b(la)h)what', replace_with='|') == 'zoom|what'

    def test_closest_index(self):
        # Basic tests
        assert closest_index([1, 2, 3, 4], 3) == 2  # Basic test
        assert closest_index([1, 2, 3], 1) == 0  # Special: The first element is the one sought
        assert closest_index(numpy.array([1, 2, 3, 4]), 4) == 3  # Special: The last element is the one sought
        assert closest_index([1, 2, 2, 3], 2) == 1  # Special: duplicated value: pick first instance
        assert closest_index([1, 2, 3, 4], 2.2) == 1  # Make sure it works for numbers in between
        assert closest_index([1, 2, 3, 4], 2.7) == 2
        # Exception handling and coping with problems
        self.assertRaises(TypeError, closest_index, 5, 5)  # First arg is not a list --> TypeError
        self.assertRaises(TypeError, closest_index, [1, 2, 3], 'string_not_number')
        self.assertRaises(TypeError, closest_index, [1, 2, 3], [3, 2, 1])  # Can't call w/ list as 2nd arg unless len=1

    def test_list_structures(self):  # Also tests dict_structures
        struct_list = list_structures(omas_rcparams['default_imas_version'])
        struct_list2 = list_structures(self.specific_test_version)
        assert isinstance(struct_list, list)
        assert isinstance(struct_list2, list)
        assert isinstance(struct_list[0], basestring)
        assert 'pf_active' in struct_list2
        struct_dict = dict_structures(omas_rcparams['default_imas_version'])
        struct_dict2 = dict_structures(self.specific_test_version)
        assert isinstance(struct_dict, dict)
        assert isinstance(struct_dict2, dict)
        assert 'pf_active' in struct_dict2.keys()
        assert all([item in struct_dict.keys() for item in struct_list])
        assert all([item in struct_dict2.keys() for item in struct_list2])

    def test_omas_info(self):
        get_list = ['pf_active', 'thomson_scattering', 'charge_exchange']
        ods_info_pfa = omas_info(get_list[0])
        ods_info_pfa2 = omas_info(get_list[0], self.specific_test_version)
        assert get_list[0] in ods_info_pfa.keys()
        assert get_list[0] in ods_info_pfa2.keys()
        if get_list[0] == 'pf_active':
            assert isinstance(ods_info_pfa['pf_active.circuit.0.connections.documentation'], basestring)
        ods_info_list = omas_info(get_list)
        assert all([item in ods_info_list for item in get_list])

    def test_o2u(self):
        assert o2u('equilibrium.time_slice.0.global_quantities.ip') == 'equilibrium.time_slice.:.global_quantities.ip'
        assert o2u('equilibrium.time_slice.:.global_quantities.ip') == 'equilibrium.time_slice.:.global_quantities.ip'
        assert o2u('2') == ':'
        assert o2u('equilibrium') == 'equilibrium'
        assert o2u('equilibrium.2') == 'equilibrium.:'

    def test_set_time_array(self):
        ods = ODS()
        ods.set_time_array('equilibrium.vacuum_toroidal_field.b0', 0, 0.1)
        ods.set_time_array('equilibrium.vacuum_toroidal_field.b0', 1, 0.2)

    def test_info(self):
        omas_info('equilibrium')
        omas_info(None)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestOmasUtils)
    unittest.TextTestRunner(verbosity=2).run(suite)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Save through data systems
=========================

This examples instantiates a sample ODS, and saves it and loads it trhough the different systems that OMAS supports.
Finally, a check is done to make sure that the final ODS is the same as the initial ODS.

.. figure:: ../images/omas_through_simple.png
  :align: center
  :width: 50%
  :alt: omas save data through different file formats
  :target: ../_images/omas_through_simple.png
"""

from __future__ import print_function, division, unicode_literals

import os

os.environ['OMAS_DEBUG_TOPIC'] = '*'

from omas import *

# load some sample data
ods_start = ods_sample()

# save/load Python pickle
filename = 'test.pkl'
save_omas_pkl(ods_start, filename)
ods = load_omas_pkl(filename)

# save/load ASCII json
filename = 'test.json'
save_omas_json(ods, filename)
ods = load_omas_json(filename)

# save/load NetCDF
filename = 'test.nc'
save_omas_nc(ods, filename)
ods = load_omas_nc(filename)

# save/load HDF5
filename = 'test.h5'
save_omas_h5(ods, filename)
ods = load_omas_h5(filename)

# remote save/load S3
filename = 'test.s3'
save_omas_s3(ods, filename)
ods = load_omas_s3(filename)

# save/load IMAS
omas_rcparams['allow_fake_imas_fallback'] = True
paths = save_omas_imas(ods, machine='ITER', pulse=1, new=True)
ods_end = load_omas_imas(machine='ITER', pulse=1, paths=paths)

# check data
check = different_ods(ods_start, ods_end)
if not check:
    print('OMAS data got saved and loaded correctly throughout')
else:
    pprint(check)



#!/usr/bin/env python
# # -*- coding: utf-8 -*-

"""
Test script for omas saving/loading data in different formats

.. code-block:: none

   python -m unittest omas/tests/test_omas_suite

-------
"""

from __future__ import print_function, division, unicode_literals
import unittest

import os
import numpy
from omas import *

try:
    import imas

    failed_IMAS = False
except ImportError as _excp:
    failed_IMAS = _excp

try:
    import hdc

    failed_HDC = False
except ImportError as _excp:
    failed_HDC = _excp

try:
    import boto3

    if not os.path.exists(os.environ.get('AWS_CONFIG_FILE', os.environ['HOME'] + '/.aws/config')):
        raise RuntimeError('Missing AWS configuration file ~/.aws/config')
    failed_S3 = False
except RuntimeError as _excp:
    failed_S3 = _excp

try:
    from pymongo import MongoClient
    from pymongo.errors import ServerSelectionTimeoutError
    from omas.omas_mongo import get_mongo_credentials

    up = get_mongo_credentials(server=omas_rcparams['default_mongo_server'])
    client = MongoClient(omas_rcparams['default_mongo_server'].format(**up), serverSelectionTimeoutMS=1000)
    client.server_info()
    failed_mongo = False
except ServerSelectionTimeoutError as _excp:
    failed_mongo = _excp


class TestOmasSuite(unittest.TestCase):

    def test_omas_pkl(self):
        ods = ods_sample()
        ods1 = through_omas_pkl(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('pkl through difference: %s' % diff)

    def test_omas_json(self):
        ods = ods_sample()
        ods1 = through_omas_json(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('json through difference: %s' % diff)

    def test_omas_nc(self):
        ods = ods_sample()
        ods1 = through_omas_nc(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('nc through difference: %s' % diff)

    def test_omas_h5(self):
        ods = ods_sample()
        ods1 = through_omas_h5(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('h5 through difference: %s' % diff)

    def test_omas_ds(self):
        ods = ods_sample()
        ods1 = through_omas_ds(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('ds through difference: %s' % diff)

    def test_omas_dx(self):
        ods = ods_sample()
        odx = ods_2_odx(ods)
        odx1 = through_omas_dx(odx)
        ods1 = odx_2_ods(odx1)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('dx through difference: %s' % diff)

    @unittest.skipUnless(not failed_mongo, str(failed_mongo))
    def test_omas_mongo(self):
        ods = ods_sample()
        ods1 = through_omas_mongo(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('mongo through difference: %s' % diff)

    @unittest.skipUnless(not failed_S3, str(failed_S3))
    def test_omas_s3(self):
        ods = ods_sample()
        ods1 = through_omas_s3(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('s3 through difference: %s' % diff)

    @unittest.skipUnless(not failed_IMAS, str(failed_IMAS))
    def test_omas_imas(self):
        ods = ods_sample()
        ods1 = through_omas_imas(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('hdc through difference: %s' % diff)

    @unittest.skipUnless(not failed_HDC, str(failed_HDC))
    def test_omas_hdc(self):
        ods = ods_sample()
        ods1 = through_omas_hdc(ods)
        diff = ods.diff(ods1)
        if diff:
            raise AssertionError('hdc through difference: %s' % diff)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestOmasSuite)
    unittest.TextTestRunner(verbosity=2).run(suite)



from __future__ import print_function, division, unicode_literals

import os, sys
sys.path.insert(0,os.path.split(os.path.split(os.path.split(os.path.abspath(__file__))[0])[0])[0])

import numpy
from matplotlib import pyplot
from pprint import pprint
from omfit.classes.omfit_eqdsk import OMFITgeqdsk
from omfit.classes.omfit_onetwo import OMFITstatefile
from omas import *

# settings
os.environ['OMAS_DEBUG_TOPIC'] = 'imas'
omas_rcparams['allow_fake_imas_fallback'] = True

# generate equilibrium sample
eq = OMFITgeqdsk(imas_json_dir + '/../samples/g145419.02100')
eq.resample(17)
eq['fluxSurfaces'].changeResolution(1)
ods = eq.to_omas()
save_omas_json(ods, imas_json_dir + '/../samples/sample_equilibrium_ods.json')

# generate core_profiles and core_sources sample
state = OMFITstatefile(imas_json_dir + '/../samples/state145419_02100.nc')
state.load()
ods = state.to_omas()

# subsample core_profiles and core_sources
coordsio = {}
coordsio['core_profiles.profiles_1d.0.grid.rho_tor_norm'] = numpy.linspace(0, 1, 11)
for k in ods['core_sources.source']:
    coordsio['core_sources.source.%d.profiles_1d.0.grid.rho_tor_norm' % k] = numpy.linspace(0, 1, 11)

# write ccore_profiles and core_sources sample
for what in ['core_sources', 'core_profiles']:
    ods_subsampled = ODS()
    with omas_environment(ods, coordsio=coordsio):
        ods_subsampled[what].update(ods[what])
    save_omas_json(ods_subsampled, imas_json_dir + '/../samples/sample_%s_ods.json' % what)



'''sample ODS methods and utilities

-------
'''

from __future__ import print_function, division, unicode_literals

from .omas_utils import *
from .omas_core import ODS
from .omas_physics import constants
from .omas_plot import geo_type_lookup

__all__ = []
__ods__ = []


def add_to_ODS(f):
    '''
    anything wrapped here will be available as a ODS method with name 'sample_'+f.__name__
    '''
    __ods__.append(f.__name__)
    __all__.append(f.__name__)
    return f


def ods_sample():
    '''
    returns an ODS populated with all of the samples

    :return: sample ods
    '''
    ods = ODS()
    for ds in __all__:
        printd('Adding %s sample data to ods' % ds, topic='sample')
        ods = eval(ds)(ods)
    return ods


@add_to_ODS
def dataset_description(ods):
    ods['dataset_description.data_entry.machine'] = 'test'
    ods['dataset_description.data_entry.pulse'] = 100
    return ods


@add_to_ODS
def equilibrium(
        ods, time_index=0, include_profiles=True, include_phi=True, include_psi=True, include_wall=True, include_q=True
):
    """
    Add sample equilibrium data

    :param ods: ODS instance

    :param time_index: int
        Under which time index should fake equilibrium data be loaded?

    :param include_profiles: bool
        Include 1D profiles of pressure, q, p', FF'

    :param include_phi: bool
        Include 1D and 2D profiles of phi (toroidal flux, for calculating rho)

    :param include_psi: bool
        Include 1D and 2D profiles of psi (poloidal flux)

    :param include_wall: bool
        Include the first wall

    :param include_q: bool
        Include safety factor

    :return: ODS instance with equilibrium data added
        Since the original is modified, it is not necessary to catch the return, but it may be convenient to do so in
        some contexts. If you do not want the original to be modified, deepcopy it first.
    """
    from omas import load_omas_json
    eq = load_omas_json(imas_json_dir + '/../samples/sample_equilibrium_ods.json', consistency_check=False)

    phi = eq['equilibrium.time_slice.0.profiles_1d.phi']
    psi = eq['equilibrium.time_slice.0.profiles_1d.psi']
    q = eq['equilibrium.time_slice.0.profiles_1d.q']

    if not include_profiles:
        del eq['equilibrium.time_slice.0.profiles_1d']

    if not include_phi:
        if 'profiles_1d' in eq['equilibrium.time_slice.0']:
            del eq['equilibrium.time_slice.0.profiles_1d.phi']
        del eq['equilibrium.time_slice.0.profiles_2d.0.phi']
    else:
        eq['equilibrium.time_slice.0.profiles_1d.phi'] = phi

    if not include_psi:
        if 'profiles_1d' in eq['equilibrium.time_slice.0'] and 'psi' in eq['equilibrium.time_slice.0.profiles_1d']:
            del eq['equilibrium.time_slice.0.profiles_1d.psi']
        del eq['equilibrium.time_slice.0.profiles_2d.0.psi']
    else:
        eq['equilibrium.time_slice.0.profiles_1d.psi'] = psi

    if not include_q:
        if 'profiles_1d' in eq['equilibrium.time_slice.0'] and 'q' in eq['equilibrium.time_slice.0.profiles_1d']:
            del eq['equilibrium.time_slice.0.profiles_1d.q']
    else:
        eq['equilibrium.time_slice.0.profiles_1d.q'] = q

    if not include_wall:
        del eq['wall']

    ods['equilibrium.time_slice'][time_index].update(eq['equilibrium.time_slice.0'])
    ods['equilibrium.time_slice'][time_index]['time'] = float(time_index)
    ods['equilibrium.vacuum_toroidal_field.r0'] = eq['equilibrium.vacuum_toroidal_field.r0']
    ods.set_time_array(
        'equilibrium.vacuum_toroidal_field.b0', time_index, eq['equilibrium.vacuum_toroidal_field.b0'][0]
    )
    ods['equilibrium.time'] = ods['equilibrium.time_slice[:].time']

    return ods


@add_to_ODS
def core_profiles(ods, time_index=0, nx=11, add_junk_ion=False, include_pressure=True):
    """
    Add sample core_profiles data

    :param ods: ODS instance

    :param time_index: int

    :param nx: int
        Number of points in test profiles

    :param add_junk_ion: bool
        Flag for adding a junk ion for testing how well functions tolerate problems. This will be missing labels, etc.

    :param include_pressure: bool
        Include pressure profiles when temperature and density are added

    :return: ODS instance with profiles added.
        Since the original is modified, it is not necessary to catch the return, but it may be convenient to do so in
        some contexts. If you do not want the original to be modified, deepcopy it first.
    """
    from omas import load_omas_json
    pr = load_omas_json(imas_json_dir + '/../samples/sample_core_profiles_ods.json', consistency_check=False)

    ods['core_profiles.profiles_1d'][time_index].update(pr['core_profiles.profiles_1d.0'])
    ods['core_profiles.vacuum_toroidal_field.r0'] = pr['core_profiles.vacuum_toroidal_field.r0']
    ods.set_time_array('core_profiles.vacuum_toroidal_field.b0', time_index, pr['core_profiles.vacuum_toroidal_field.b0'][0])

    if add_junk_ion:
        ions = ods['core_profiles.profiles_1d'][time_index]['ion']
        ions[len(ions)] = copy.deepcopy(ions[len(ions) - 1])
        for item in ions[len(ions) - 1].flat():
            ions[len(ions) - 1][item] *= 0

    if not include_pressure:
        for item in ods.physics_core_profiles_pressures(update=False).flat().keys():
            if 'pres' in item and item in ods:
                del ods[item]

    ods['core_profiles.profiles_1d'][time_index]['time'] = float(time_index)

    return ods


@add_to_ODS
def pf_active(ods, nc_weird=0, nc_undefined=0):
    """
    Adds some FAKE active PF coil locations so that the overlay plot will work in tests.
    It's fine to test with dummy data as long as you know it's not real.

    :param ods: ODS instance

    :param nc_weird: int
        Number of coils with badly defined geometry to include for testing plot overlay robustness

    :param nc_undefined: int
        Number of coils with undefined geometry_type (But valid r, z outlines) to include for testing plot overlay
        robustness.

    :return: ODS instance with FAKE PF ACTIVE HARDWARE INFORMATION added.
    """

    nc_reg = 4
    nc = nc_reg + nc_weird + nc_undefined
    fc_dat = [
        #  R        Z       dR      dZ    tilt1  tilt2
        [.8608, .16830, .0508, .32106, 0.0, 0.0],
        [1.0041, 1.5169, .13920, .11940, 45.0, 0.0],
        [2.6124, 0.4376, 0.17320, 0.1946, 0.0, 92.40],
        [2.3834, -1.1171, 0.1880, 0.16920, 0.0, -108.06],
    ]

    rect_code = geo_type_lookup('rectangle', 'pf_active', ods.imas_version, reverse=True)
    outline_code = geo_type_lookup('outline', 'pf_active', ods.imas_version, reverse=True)

    for i in range(nc_reg):
        if (fc_dat[i][4] == 0) and (fc_dat[i][5] == 0):
            rect = ods['pf_active.coil'][i]['element.0.geometry.rectangle']
            rect['r'] = fc_dat[i][0]
            rect['z'] = fc_dat[i][1]
            rect['width'] = fc_dat[i][2]  # Or width in R
            rect['height'] = fc_dat[i][3]  # Or height in Z
            ods['pf_active.coil'][i]['element.0.geometry.geometry_type'] = rect_code
        else:
            outline = ods['pf_active.coil'][i]['element.0.geometry.outline']
            fdat = fc_dat[i]
            fdat[4] = -fc_dat[i][4] * numpy.pi / 180.
            fdat[5] = -(fc_dat[i][5] * numpy.pi / 180. if fc_dat[i][5] != 0 else numpy.pi / 2.)
            outline['r'] = [
                fdat[0] - fdat[2] / 2. - fdat[3] / 2. * numpy.tan((numpy.pi / 2. + fdat[5])),
                fdat[0] - fdat[2] / 2. + fdat[3] / 2. * numpy.tan((numpy.pi / 2. + fdat[5])),
                fdat[0] + fdat[2] / 2. + fdat[3] / 2. * numpy.tan((numpy.pi / 2. + fdat[5])),
                fdat[0] + fdat[2] / 2. - fdat[3] / 2. * numpy.tan((numpy.pi / 2. + fdat[5])),
            ]
            outline['z'] = [
                fdat[1] - fdat[3] / 2. - fdat[2] / 2. * numpy.tan(-fdat[4]),
                fdat[1] + fdat[3] / 2. - fdat[2] / 2. * numpy.tan(-fdat[4]),
                fdat[1] + fdat[3] / 2. + fdat[2] / 2. * numpy.tan(-fdat[4]),
                fdat[1] - fdat[3] / 2. + fdat[2] / 2. * numpy.tan(-fdat[4]),
            ]
            ods['pf_active.coil'][i]['element.0.geometry.geometry_type'] = outline_code

    for i in range(nc_reg, nc_reg + nc_weird):
        # This isn't a real geometry_type, so it should trigger the contingency
        ods['pf_active.coil'][i]['element.0.geometry.geometry_type'] = 99
    for i in range(nc_reg + nc_weird, nc):
        # This one doesn't have geometry_type defined, so the plot overlay will have trouble looking up which type it is
        outline = ods['pf_active.coil'][i]['element.0.geometry.outline']
        outline['r'] = [1.5, 1.6, 1.7, 1.5]
        outline['z'] = [0.1, 0.3, -0.1, 0]

    # generate some data that is not time homogeneous
    for i in range(nc_reg):
        n = (1 + i) * 10
        outline = ods['pf_active.coil'][i]['current.data'] = numpy.linspace(0, 1, n)
        outline = ods['pf_active.coil'][i]['current.time'] = numpy.linspace(0, 1, n)

    return ods


@add_to_ODS
def magnetics(ods):
    """
    Adds some FAKE magnetic probe locations so that the overlay plot will work in tests. It's fine to test
    with dummy data as long as you know it's not real.

    :param ods: ODS instance

    :return: ODS instance with FAKE MAGNETICS HARDWARE INFORMATION added.
    """

    nbp = 12
    nfl = 7

    r0 = 1.5
    z0 = 0.0
    abp = 0.8
    afl = 1.0

    angle_bp = numpy.linspace(0, 2 * numpy.pi, nbp + 1)[:-1]
    rp = r0 + abp * numpy.cos(angle_bp)
    zp = z0 + abp * numpy.sin(angle_bp)

    angle_fl = numpy.linspace(0, 2 * numpy.pi, nfl + 1)[:-1]
    rf = r0 + afl * numpy.cos(angle_fl)
    zf = z0 + afl * numpy.sin(angle_fl)

    for i in range(nbp):
        ods['magnetics.b_field_pol_probe'][i]['identifier'] = 'FAKE bpol probe {}'.format(i)
        ods['magnetics.b_field_pol_probe'][i]['position.r'] = rp[i]
        ods['magnetics.b_field_pol_probe'][i]['position.z'] = zp[i]
        ods['magnetics.b_field_pol_probe'][i]['position.phi'] = 6.5

    for i in range(nfl):
        ods['magnetics.flux_loop'][i]['identifier'] = 'FAKE flux loop {}'.format(i)
        ods['magnetics.flux_loop'][i]['position.0.r'] = rf[i]
        ods['magnetics.flux_loop'][i]['position.0.z'] = zf[i]

    ods['magnetics.time'] = [0]

    return ods


@add_to_ODS
def thomson_scattering(ods, nc=10):
    """
    Adds some FAKE Thomson scattering channel locations so that the overlay plot will work in tests. It's fine to test
    with dummy data as long as you know it's not real.

    :param ods: ODS instance

    :param nc: Number of channels to add.

    :return: ODS instance with FAKE THOMSON HARDWARE INFORMATION added.
    """

    r = numpy.linspace(1.935, 1.945, nc)
    z = numpy.linspace(-0.7, 0.2, nc)
    for i in range(nc):
        ch = ods['thomson_scattering.channel'][i]
        ch['identifier'] = 'F_TS_{:02d}'.format(i)  # F for fake
        ch['name'] = 'Fake Thomson channel for testing {}'.format(i)
        ch['position.phi'] = 6.5  # This angle in rad should look bad to someone who doesn't notice the Fake labels
        ch['position.r'] = r[i]
        ch['position.z'] = z[i]

    ods['thomson_scattering.time'] = [0]

    return ods


@add_to_ODS
def charge_exchange(ods, nc=10):
    """
    Adds some FAKE CER channel locations so that the overlay plot will work in tests. It's fine to test
    with dummy data as long as you know it's not real. This function can overwrite existing data if you're not careful.
    The original is modified, so deepcopy first if you want different ODSs.

    :param ods: ODS instance

    :param nc: Number of channels to add.

    :return: ODS instance with FAKE CER HARDWARE INFORMATION added.
    """

    r = numpy.linspace(1.4, 2.2, nc)
    z = numpy.linspace(0.05, -0.07, nc)
    for i in range(nc):
        ch = ods['charge_exchange.channel'][i]
        ch['identifier'] = 'FAKE_CER_{:02d}'.format(i)  # F for fake
        ch['name'] = 'Fake CER channel for testing {}'.format(i)
        for x in ['r', 'z', 'phi']:
            ch['position'][x]['time'] = numpy.array([0])
        ch['position.phi.data'] = numpy.array([6.5])
        ch['position.r.data'] = numpy.array([r[i]])
        ch['position.z.data'] = numpy.array([z[i]])

    return ods


@add_to_ODS
def interferometer(ods):
    """
    Adds some FAKE interferometer locations so that the overlay plot will work in tests. It's fine to test
    with dummy data as long as you know it's not real.

    :param ods: ODS instance

    :return: ODS instance with FAKE INTERFEROMETER HARDWARE INFORMATION added.
    """
    ods['interferometer.channel.0.identifier'] = 'FAKE horz. interf.'
    r0 = ods['interferometer.channel.0.line_of_sight']
    r0['first_point.phi'] = r0['second_point.phi'] = 225 * (-numpy.pi / 180)
    r0['first_point.r'], r0['second_point.r'] = 3.0, 0.8
    r0['first_point.z'] = r0['second_point.z'] = 0.0

    i = 0
    ods['interferometer.channel'][i + 1]['identifier'] = 'FAKE vert. interf.'
    los = ods['interferometer.channel'][i + 1]['line_of_sight']
    los['first_point.phi'] = los['second_point.phi'] = 240 * (-numpy.pi / 180)
    los['first_point.r'] = los['second_point.r'] = 1.48
    los['first_point.z'], los['second_point.z'] = -1.8, 1.8

    for j in range(len(ods['interferometer.channel'])):
        ch = ods['interferometer.channel'][j]
        ch['line_of_sight.third_point'] = copy.deepcopy(ch['line_of_sight.first_point'])

    ods['interferometer.time'] = [0]

    return ods


@add_to_ODS
def bolometer(ods, nc=10):
    """
    Adds some FAKE bolometer chord locations so that the overlay plot will work in tests. It's fine to test
    with dummy data as long as you know it's not real.

    :param ods: ODS instance

    :param nc: 10  # Number of fake channels to make up for testing

    :return: ODS instance with FAKE BOLOMETER HARDWARE INFORMATION added.
    """

    angles = numpy.pi + numpy.linspace(-numpy.pi / 4.0, numpy.pi / 4.0, nc)

    # FAKE origin for the FAKE bolometer fan
    r0 = 2.5
    z0 = 0.05

    for i in range(nc):
        ch = ods['bolometer.channel'][i]['line_of_sight']
        ch['first_point.r'] = r0
        ch['first_point.z'] = z0 + 0.001 * i
        ch['second_point.r'] = ch['first_point.r'] + numpy.cos(angles[i])
        ch['second_point.z'] = ch['first_point.z'] + numpy.sin(angles[i])
        ods['bolometer.channel'][i]['identifier'] = 'fake bolo {}'.format(i)

    ods['bolometer.channel'][nc - 1]['identifier'] = 'bolo fan 2 fake'  # This tests separate colors per fan in overlay

    ods['bolometer.time'] = [0]

    return ods


@add_to_ODS
def gas_injection(ods):
    """
    Adds some FAKE gas injection locations so that the overlay plot will work in tests. It's fine to test
    with dummy data as long as you know it's not real. This function can overwrite existing data if you're not careful.
    The original is modified, so deepcopy first if you want different ODSs.

    :param ods: ODS instance

    :return: ODS instance with FAKE GAS INJECTION HARDWARE INFORMATION added.
    """

    ods['gas_injection.time'] = [0]

    ods['gas_injection.pipe.0.name'] = 'FAKE_GAS_A'
    ods['gas_injection.pipe.0.exit_position.r'] = 2.25
    ods['gas_injection.pipe.0.exit_position.z'] = 0.0
    ods['gas_injection.pipe.0.exit_position.phi'] = 6.5
    ods['gas_injection.pipe.0.valve.0.identifier'] = 'FAKE_GAS_VALVE_A'

    ods['gas_injection.pipe.1.name'] = 'FAKE_GAS_B'
    ods['gas_injection.pipe.1.exit_position.r'] = 1.65
    ods['gas_injection.pipe.1.exit_position.z'] = 1.1
    ods['gas_injection.pipe.1.exit_position.phi'] = 6.5
    ods['gas_injection.pipe.1.valve.0.identifier'] = 'FAKE_GAS_VALVE_B'
    ods['gas_injection.pipe.1.second_point.r'] = 1.63
    ods['gas_injection.pipe.1.second_point.z'] = 1.08
    ods['gas_injection.pipe.1.second_point.phi'] = 6.5

    ods['gas_injection.pipe.2.name'] = 'FAKE_GAS_C'
    ods['gas_injection.pipe.2.exit_position.r'] = 2.1
    ods['gas_injection.pipe.2.exit_position.z'] = -0.6
    ods['gas_injection.pipe.2.valve.0.identifier'] = 'FAKE_GAS_VALVE_C'
    # This one deliberately doesn't have a phi angle defined, for testing purposes.

    return ods


@add_to_ODS
def langmuir_probes(ods):
    """
    Adds some FAKE Langmuir probe locations so the overlay plot will work in tests.
    :param ods: ODS instance

    :return: ODS instance
        The data are also written to the original, so you don't have to catch the return
    """

    ods['langmuir_probes.time'] = numpy.array([0])

    ods['langmuir_probes.embedded.0.identifier'] = 0
    ods['langmuir_probes.embedded.0.name'] = 'p1'
    ods['langmuir_probes.embedded.0.position.r'] = 0.9
    ods['langmuir_probes.embedded.0.position.z'] = 0.1
    ods['langmuir_probes.embedded.0.position.phi'] = 0

    ods['langmuir_probes.embedded.1.identifier'] = -1
    ods['langmuir_probes.embedded.1.name'] = 'p23'
    ods['langmuir_probes.embedded.1.position.r'] = 0.9
    ods['langmuir_probes.embedded.1.position.z'] = -0.9
    ods['langmuir_probes.embedded.1.position.phi'] = numpy.NaN

    ods['langmuir_probes.embedded.2.identifier'] = -2
    ods['langmuir_probes.embedded.2.name'] = 'blah'
    ods['langmuir_probes.embedded.2.position.r'] = 1.5
    ods['langmuir_probes.embedded.2.position.z'] = -1.25
    ods['langmuir_probes.embedded.2.position.phi'] = numpy.NaN

    ods['langmuir_probes.embedded.3.identifier'] = -3
    ods['langmuir_probes.embedded.3.name'] = 'donkey!'
    ods['langmuir_probes.embedded.3.position.r'] = 1.525
    ods['langmuir_probes.embedded.3.position.z'] = -1.25
    ods['langmuir_probes.embedded.3.position.phi'] = numpy.NaN

    ods['langmuir_probes.embedded.4.identifier'] = -4
    ods['langmuir_probes.embedded.4.name'] = 'zzz'
    ods['langmuir_probes.embedded.4.position.r'] = 1.4
    ods['langmuir_probes.embedded.4.position.z'] = 1.4
    ods['langmuir_probes.embedded.4.position.phi'] = numpy.pi

    ods['langmuir_probes.embedded.5.identifier'] = -5
    ods['langmuir_probes.embedded.5.name'] = "it's just a test"
    ods['langmuir_probes.embedded.5.position.r'] = 2.45
    ods['langmuir_probes.embedded.5.position.z'] = 0.25
    ods['langmuir_probes.embedded.5.position.phi'] = numpy.NaN

    return ods


@add_to_ODS
def wall(ods):
    """
    Adds some FAKE wall data to support testing. The wall is approximately DIII-D shaped, but I didn't try very hard.

    :param ods: ODS instance

    :return: ODS instance
        Edits are done in-place, so you don't have to catch the return (but you can if you want to!)
    """
    ods['wall.description_2d[0].limiter.type.description'] = 'first wall'
    ods['wall.description_2d[0].limiter.type.index'] = 0
    ods['wall.description_2d[0].limiter.type.name'] = 'first_wall'
    ods['wall.description_2d[0].limiter.unit[0].outline.r'] = [
        1.0, 1.0, 1.3, 1.4, 1.6, 2.15, 2.35, 2.35, 2.15, 1.800, 1.350, 1.35, 1.10, 1.00, 1.0
    ]
    ods['wall.description_2d[0].limiter.unit[0].outline.z'] = [
        0.0, 1.4, 1.4, 1.3, 1.1, 1.00, 0.50, -0.5, -1.0, -1.25, -1.25, -1.4, -1.4, -1.3, 0.0
    ]

    return ods



#!/usr/bin/env python
# # -*- coding: utf-8 -*-
"""
Interface with GKDB
===================
Use OMAS to interface with Gyro-Kinetic DataBase (GKDB) https://gitlab.com/gkdb/gkdb
GKDB is a publicly accessible database of delta-f flux-tube gyro-kinetic simulations of tokamak plasmas
which stores its data according to the `gyrokinetic` IMAS IDS https://gafusion.github.io/omas/schema/schema_gyrokinetics.html
"""

from omas import ODS, imas_json_dir
from pprint import pprint
import tempfile
import sys

# load a sample GKDB sample json file
# note use of `consistency_check='warn'` is necessary since GKDB json is in fact not yet 100% compatible with IMAS
sample_filename = imas_json_dir + '/../samples/gkdb_linear_eigenvalue.json'
ods = ODS(consistency_check='warn')
ods['gyrokinetics'].load(sample_filename)

# show content
pprint(ods.pretty_paths())

# save a copy
filename = tempfile.gettempdir() + '/gkdb_linear_initialvalue.json'
ods['gyrokinetics'].save(filename)

# load the newly saved copy
ods1 = ODS()
ods1['gyrokinetics'].load(filename, consistency_check='warn')

# look for differences between original GKDB json and OMAS json
differences = ods.diff(ods1, ignore_type=True)
if not differences:
    print('\nPrint no differences found: save/load of GKDB json file worked\n')
else:
    pprint(differences)
    raise RuntimeError('Save/Load of GKDB on json file failed')

# raise error if trying to run GKDB under Python2x
try:
    if sys.version_info < (3, 0):
        raise ImportError('gkdb library is only Python 3 compatible')
    import gkdb.core.model
except ImportError as _excp:
    print('Could not import gkdb library: %s' % repr(_excp))
else:
    # Check that GKDB file written by OMAS is valid also according to GKDB
    if gkdb.core.ids_checks.check_json(filename, only_input=False):
        print('json file saved via OMAS is valid for gkdb')

    # This requires an account on the GKDB server
    if False:
        gkdb.core.model.connect_to_gkdb()
        gkdb.core.model.Ids_properties.from_json(filename)



'''
Utility to generate the omas/omas_cocos.py file
'''

import os, sys
sys.path.insert(0,os.path.split(os.path.split(os.path.split(os.path.abspath(__file__))[0])[0])[0])

from omas.omas_utils import list_structures, omas_rcparams
from omas.omas_physics import generate_cocos_signals

generate_cocos_signals(list_structures(imas_version=omas_rcparams['default_imas_version']), threshold=0, write=True)



#!/usr/bin/env python
# # -*- coding: utf-8 -*-
"""
Test all save/load systems
==========================
Save and load a sample ODS through the different OMAS save formats and check that information does not get lost or corrupted.
A final report summarizes if some storage systems combinations have errors.
"""

from __future__ import print_function, division, unicode_literals

import os
import numpy
from omas import *
from omas.omas_utils import printd

_tests = ['pkl', 'json', 'nc', 's3', 'h5', 'imas', 'hdc']


def through_omas_suite(ods=None, test_type=None, do_raise=False):
    """
    :param ods: omas structure to test. If None this is set to ods_sample

    :param test_type: None tests all suite, otherwise choose among %s

    :param do_raise: raise error if something goes wrong
    """

    if ods is None:
        ods = ods_sample()

    if test_type in _tests:
        os.environ['OMAS_DEBUG_TOPIC'] = test_type
        ods1 = globals()['through_omas_' + test_type](ods)
        check = different_ods(ods, ods1)
        if not check:
            print('OMAS data got saved and loaded correctly')
        else:
            pprint(check)

    else:
        os.environ['OMAS_DEBUG_TOPIC'] = '*'
        printd('OMAS is using IMAS data structure version `%s` as default' % omas_rcparams['default_imas_version'], topic='*')

        print('=' * 20)

        results = numpy.zeros((len(_tests), len(_tests)))

        for k1, t1 in enumerate(_tests):
            failed1 = False
            try:
                ods1 = globals()['through_omas_' + t1](ods)
            except Exception as _excp:
                failed1 = _excp
                if do_raise:
                    raise
            for k2, t2 in enumerate(_tests):
                try:
                    if failed1:
                        raise failed1
                    ods2 = globals()['through_omas_' + t2](ods1)

                    different = different_ods(ods1, ods2)
                    if not different:
                        print('FROM %s TO %s : OK' % (t1.center(5), t2.center(5)))
                        results[k1, k2] = 1.0
                    else:
                        print('FROM %s TO %s : NO --> %s' %
                              (t1.center(5), t2.center(5), different))
                        results[k1, k2] = -1.0

                except Exception as _excp:
                    print('FROM %s TO %s : NO --> %s' %
                          (t1.center(5), t2.center(5), repr(_excp)))
                    if do_raise:
                        raise
        print('=' * 20)
        print(results.astype(int))
        print('=' * 20)


through_omas_suite(ods=None, test_type=None, do_raise=False)

###################################
# In the above example `consistency_check = True` result in the following error::
#
#     LookupError: Not a valid IMAS 3.18.0 location: `equilibrium.time_slice.:.does_not_exist`
#                                                                              ^^^^^^^^^^^^^^
#     Did you mean: ['profiles_2d', 'ggd', 'boundary', 'profiles_1d', 'constraints', 'global_quantities', 'coordinate_system', 'boundary_separatrix', 'time', 'convergence']

#################################################
# If successful, the test should output:
#
# .. code-block:: none
#
#     ====================
#     FROM  pkl  TO  pkl  : OK
#     FROM  pkl  TO  json : OK
#     FROM  pkl  TO   nc  : OK
#     FROM  pkl  TO   s3  : OK
#     FROM  pkl  TO  imas : OK
#     FROM  json TO  pkl  : OK
#     FROM  json TO  json : OK
#     FROM  json TO   nc  : OK
#     FROM  json TO   s3  : OK
#     FROM  json TO  imas : OK
#     FROM   nc  TO  pkl  : OK
#     FROM   nc  TO  json : OK
#     FROM   nc  TO   nc  : OK
#     FROM   nc  TO   s3  : OK
#     FROM   nc  TO  imas : OK
#     FROM   s3  TO  pkl  : OK
#     FROM   s3  TO  json : OK
#     FROM   s3  TO   nc  : OK
#     FROM   s3  TO   s3  : OK
#     FROM   s3  TO  imas : OK
#     FROM  imas TO  pkl  : OK
#     FROM  imas TO  json : OK
#     FROM  imas TO   nc  : OK
#     FROM  imas TO   s3  : OK
#     FROM  imas TO  imas : OK
#     ====================
#     [[1 1 1 1 1]
#      [1 1 1 1 1]
#      [1 1 1 1 1]
#      [1 1 1 1 1]
#      [1 1 1 1 1]]
#     ====================
#
# If for example `imas` is not installed on the system, the test output will look like:
#
# .. code-block:: none
#
#     ====================
#     FROM  pkl  TO  pkl  : OK
#     FROM  pkl  TO  json : OK
#     FROM  pkl  TO   nc  : OK
#     FROM  pkl  TO   s3  : OK
#     FROM  pkl  TO  imas : NO --> ImportError('No module named imas',)
#     FROM  json TO  pkl  : OK
#     FROM  json TO  json : OK
#     FROM  json TO   nc  : OK
#     FROM  json TO   s3  : OK
#     FROM  json TO  imas : NO --> ImportError('No module named imas',)
#     FROM   nc  TO  pkl  : OK
#     FROM   nc  TO  json : OK
#     FROM   nc  TO   nc  : OK
#     FROM   nc  TO   s3  : OK
#     FROM   nc  TO  imas : NO --> ImportError('No module named imas',)
#     FROM   s3  TO  pkl  : OK
#     FROM   s3  TO  json : OK
#     FROM   s3  TO   nc  : OK
#     FROM   s3  TO   s3  : OK
#     FROM   s3  TO  imas : NO --> ImportError('No module named imas',)
#     FROM  imas TO  pkl  : NO --> ImportError('No module named imas',)
#     FROM  imas TO  json : NO --> ImportError('No module named imas',)
#     FROM  imas TO   nc  : NO --> ImportError('No module named imas',)
#     FROM  imas TO   s3  : NO --> ImportError('No module named imas',)
#     FROM  imas TO  imas : NO --> ImportError('No module named imas',)
#     ====================
#     [[1 1 1 1 0]
#      [1 1 1 1 0]
#      [1 1 1 1 0]
#      [1 1 1 1 0]
#      [0 0 0 0 0]]
#     ====================



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Interpolate whole ODS
=====================
Seamless handling of coordinates within OMAS makes it easy to reinterpolate a whole ODS on a new grid
"""

from __future__ import print_function, division, unicode_literals

import numpy
from omas import *

# original ODS
ods = ODS()
ods.sample_equilibrium()

# interpolated ODS
ods_interpolated = ODS()

# define new psi grid of ods_interpolated
new_psi = numpy.linspace(ods['equilibrium.time_slice.0.profiles_1d.psi'][0], ods['equilibrium.time_slice.0.profiles_1d.psi'][-1], 21)
ods_interpolated['equilibrium.time_slice.0.profiles_1d.psi'] = new_psi

# interpolate whole ods on new psi grid
with omas_environment(ods_interpolated, coordsio=ods):
    ods_interpolated.update(ods)

# print some quantity from interpolated ods
assert len(ods_interpolated['equilibrium.time_slice.0.profiles_1d.pressure']) == 21
print(ods_interpolated['equilibrium.time_slice.0.profiles_1d.pressure'])



'''save/load from IMAS routines

-------
'''

from __future__ import print_function, division, unicode_literals

from .omas_utils import *
from .omas_core import ODS, codeparams_xml_save, codeparams_xml_load


# --------------------------------------------
# IMAS convenience functions
# --------------------------------------------
def imas_open(user, machine, pulse, run, new=False,
              imas_version=os.environ.get('IMAS_VERSION', omas_rcparams['default_imas_version'])):
    """
    function to open an IMAS

    :param user: IMAS username

    :param machine: IMAS machine

    :param pulse: IMAS pulse

    :param run: IMAS run id

    :param new: whether the open should create a new IMAS tree

    :param imas_version: IMAS version

    :return: IMAS ids
    """
    import imas
    printd("ids = imas.ids(%d,%d)" % (pulse, run), topic='imas_code')
    ids = imas.ids(pulse, run)

    if user is None and machine is None:
        pass
    elif user is None or machine is None:
        raise Exception('user={user}, machine={machine}, imas_version={imas_version}\n'
                        'Either specify all or none of `user`, `machine`, `imas_version`\n'
                        'If none of them are specified then use `imasdb` command to set '
                        'MDSPLUS_TREE_BASE_? environmental variables'.format(user=user, machine=machine, pulse=pulse,
                                                                             run=run, imas_version=imas_version))

    if user is None and machine is None:
        if new:
            printd("ids.create()", topic='imas_code')
            ids.create()
        else:
            printd("ids.open()", topic='imas_code')
            try:
                ids.open()
            except Exception as _excp:
                if 'Error opening imas pulse' in str(_excp):
                    raise IOError('Error opening imas pulse %d run %d' % (pulse, run))
        if not ids.isConnected():
            raise Exception('Failed to establish connection to IMAS database '
                            '(pulse:{pulse} run:{run}, DB:{db})'.format(pulse=pulse, run=run, db=os.environ.get('MDSPLUS_TREE_BASE_0', '???')[:-2]))

    else:
        imas_major_version = '3'
        if new:
            printd("ids.create_env(%s, %s, %s)" % (repr(user), repr(machine), repr(imas_major_version)), topic='imas_code')
            ids.create_env(user, machine, imas_version)
        else:
            printd("ids.open_env(%s, %s, %s)" % (repr(user), repr(machine), repr(imas_major_version)), topic='imas_code')
            try:
                ids.open_env(user, machine, imas_major_version)
            except Exception as _excp:
                if 'Error opening imas pulse' in str(_excp):
                    raise IOError('Error opening imas pulse (user:%s machine:%s pulse:%s run:%s, imas_version:%s)' % (user, machine, pulse, run, imas_version))
        if not ids.isConnected():
            raise Exception('Failed to establish connection to IMAS database (user:%s machine:%s pulse:%s run:%s, imas_version:%s)' % (user, machine, pulse, run, imas_version))
    return ids


def imas_set(ids, path, value, skip_missing_nodes=False, allocate=False):
    """
    assign a value to a path of an open IMAS ids

    :param ids: open IMAS ids to write to

    :param path: ODS path

    :param value: value to assign

    :param skip_missing_nodes:  if the IMAS path does not exists:
                             `False` raise an error
                             `True` does not raise error
                             `None` prints a warning message

    :param allocate: whether to perform only IMAS memory allocation (ids.resize)

    :return: path if set was done, otherwise None
    """
    if numpy.atleast_1d(is_uncertain(value)).any():
        path = copy.deepcopy(path)
        tmp = imas_set(ids, path, nominal_values(value), skip_missing_nodes=skip_missing_nodes, allocate=allocate)
        path[-1] = path[-1] + '_error_upper'
        imas_set(ids, path, std_devs(value), skip_missing_nodes=skip_missing_nodes, allocate=allocate)
        return tmp

    ds = path[0]
    path = path[1:]

    # `info` IDS is used by OMAS to hold user, machine, pulse, run, imas_version
    # for saving methods that do not carry that information. IMAS does not store
    # this information as part of the data dictionary.
    if ds in add_datastructures.keys():
        return

    # identify data dictionary to use, from this point on `m` points to the IDS
    debug_path = ''
    if hasattr(ids, ds):
        debug_path += 'ids.%s' % ds
        m = getattr(ids, ds)
        if hasattr(m, 'time') and not isinstance(m.time, float) and not m.time.size:
            m.time.resize(1)
            m.time[0] = -1.0
    elif skip_missing_nodes is not False:
        if skip_missing_nodes is None:
            printe('WARNING: %s is not part of IMAS' % l2i([ds] + path))
        return
    else:
        printd(debug_path, topic='imas_code')
        raise AttributeError('%s is not part of IMAS' % l2i([ds] + path))

    # traverse IMAS structure until reaching the leaf
    out = m
    for kp, p in enumerate(path):
        location = l2i([ds] + path[:kp + 1])
        if isinstance(p, basestring):
            if hasattr(out, p):
                if kp < (len(path) - 1):
                    debug_path += '.' + p
                    out = getattr(out, p)
            elif skip_missing_nodes is not False:
                if skip_missing_nodes is None:
                    printe('WARNING: %s is not part of IMAS' % location)
                return
            else:
                printd(debug_path, topic='imas_code')
                raise AttributeError('%s is not part of IMAS' % location)
        else:
            try:
                out = out[p]
                debug_path += '[%d]' % p
            except IndexError:
                if not allocate:
                    raise IndexError('%s structure array exceed allocation' % location)
                printd(debug_path + ".resize(%d)" % (p + 1), topic='imas_code')
                out.resize(p + 1)
                debug_path += '[%d]' % p
                out = out[p]

    # if we are allocating data, simply stop here
    if allocate:
        return [ds] + path

    # assign data to leaf node
    printd('setting  : %s' % location, topic='imas')
    setattr(out, path[-1], value)
    if 'imas_code' in os.environ.get('OMAS_DEBUG_TOPIC', ''):  # use if statement here to avoid unecessary repr(value) when not debugging
        printd(debug_path + '.%s=%s' % (path[-1], repr(value).replace('\\n', '\n')), topic='imas_code')

    # return path
    return [ds] + path


def imas_empty(value):
    '''
    Check if value is an IMAS empty
        * array with no size
        * float of value -9E40
        * integer of value -999999999
        * empty string

    :param value: value to check

    :return: None if value is an IMAS empty
    '''
    if isinstance(value, numpy.ndarray) and not value.size:
        value = None
    # missing floats and integers
    elif (isinstance(value, float) and value == -9E40) or (isinstance(value, int) and value == -999999999):
        value = None
    # empty strings
    elif isinstance(value, basestring) and not len(value):
        value = None
    return value


def imas_get(ids, path, skip_missing_nodes=False):
    """
    read the value of a path in an open IMAS ids

    :param ids: open IMAS ids to read from

    :param path: ODS path

    :param skip_missing_nodes:  if the IMAS path does not exists:
                             `False` raise an error
                             `True` does not raise error
                             `None` prints a warning message

    :return: the value that was read if successful or None otherwise
    """
    printd('fetching: %s' % l2i(path), topic='imas')
    ds = path[0]
    path = path[1:]

    debug_path = ''
    if hasattr(ids, ds):
        debug_path += 'ids.%s' % ds
        m = getattr(ids, ds)
    elif skip_missing_nodes is not False:
        if skip_missing_nodes is None:
            printe('WARNING: %s is not part of IMAS' % l2i([ds] + path))
        return None
    else:
        printd(debug_path, topic='imas_code')
        raise AttributeError('%s is not part of IMAS' % l2i([ds] + path))

    # traverse the IDS to get the data
    out = m
    for kp, p in enumerate(path):
        if isinstance(p, basestring):
            if hasattr(out, p):
                debug_path += '.%s' % p
                out = getattr(out, p)
            elif skip_missing_nodes is not False:
                if skip_missing_nodes is None:
                    printe('WARNING: %s is not part of IMAS' % l2i([ds] + path[:kp + 1]))
                    printe(out.__dict__.keys())
                return None
            else:
                printd(debug_path, topic='imas_code')
                raise AttributeError('%s is not part of IMAS' % l2i([ds] + path[:kp + 1]))
        else:
            debug_path += '[%s]' % p
            out = out[p]

    # handle missing data
    data = imas_empty(out)

    printd(debug_path, topic='imas_code')
    return data


# --------------------------------------------
# save and load OMAS to IMAS
# --------------------------------------------
@codeparams_xml_save
def save_omas_imas(ods, user=None, machine=None, pulse=None, run=None, new=False,
                   imas_version=os.environ.get('IMAS_VERSION', omas_rcparams['default_imas_version'])):
    """
    Save OMAS data to IMAS

    :param ods: OMAS data set

    :param user: IMAS username (reads ods['dataset_description.data_entry.user'] if user is None and finally fallsback on os.environ['USER'])

    :param machine: IMAS machine (reads ods['dataset_description.data_entry.machine'] if machine is None)

    :param pulse: IMAS pulse (reads ods['dataset_description.data_entry.pulse'] if pulse is None)

    :param run: IMAS run (reads ods['dataset_description.data_entry.run'] if run is None and finally fallsback on 0)

    :param new: whether the open should create a new IMAS tree

    :param imas_version: IMAS version

    :return: paths that have been written to IMAS
    """

    # handle default values for user, machine, pulse, run, imas_version
    # it tries to re-use existing information
    if user is None:
        user = ods.get('dataset_description.data_entry.user', os.environ['USER'])
    if machine is None:
        machine = ods.get('dataset_description.data_entry.machine', None)
    if pulse is None:
        pulse = ods.get('dataset_description.data_entry.pulse', None)
    if run is None:
        run = ods.get('dataset_description.data_entry.run', 0)

    # set dataset_description entries that were empty
    if user is not None and 'dataset_description.data_entry.user' not in ods:
        ods['dataset_description.data_entry.user'] = user
    if machine is not None and 'dataset_description.data_entry.machine' not in ods:
        ods['dataset_description.data_entry.machine'] = machine
    if pulse is not None and 'dataset_description.data_entry.pulse' not in ods:
        ods['dataset_description.data_entry.pulse'] = pulse
    if run is not None and 'dataset_description.data_entry.run' not in ods:
        ods['dataset_description.data_entry.run'] = run
    if imas_version is not None and 'dataset_description.imas_version' not in ods:
        ods['dataset_description.imas_version'] = imas_version

    if user is not None and machine is not None:
        printd('Saving to IMAS (user:%s machine:%s pulse:%d run:%d, imas_version:%s)' % (user, machine, pulse, run, imas_version), topic='imas')
    elif user is None and machine is None:
        printd('Saving to IMAS (pulse:%d run:%d, DB:%s)' % (pulse, run, os.environ.get('MDSPLUS_TREE_BASE_0', '???')[:-2]), topic='imas')

    # ensure requirements for writing data to IMAS are satisfied
    ods.satisfy_imas_requirements()

    # get the list of paths from ODS
    paths = set_paths = ods.paths()

    try:
        # open IMAS tree
        ids = imas_open(user=user, machine=machine, pulse=pulse, run=run, new=new, imas_version=imas_version)

    except IOError as _excp:
        raise IOError(str(_excp) + '\nIf this is a new pulse/run then set `new=True`')

    except ImportError:
        # fallback on saving IMAS as NC file if IMAS is not installed
        if not omas_rcparams['allow_fake_imas_fallback']:
            raise
        filename = os.sep.join([omas_rcparams['fake_imas_dir'], '%s_%s_%d_%d_v%s.pkl' % (user, machine, pulse, run, imas_versions.get(imas_version, imas_version))])
        printe('Overloaded save_omas_imas: %s' % filename)
        from . import save_omas_pkl
        if not os.path.exists(omas_rcparams['fake_imas_dir']):
            os.makedirs(omas_rcparams['fake_imas_dir'])
        ods['dataset_description.data_entry.user'] = unicode(user)
        ods['dataset_description.data_entry.machine'] = unicode(machine)
        ods['dataset_description.data_entry.pulse'] = int(pulse)
        ods['dataset_description.data_entry.run'] = int(run)
        ods['dataset_description.imas_version'] = unicode(imas_version)
        save_omas_pkl(ods, filename)

    else:

        try:
            # allocate memory
            # NOTE: for how memory allocation works it is important to traverse the tree in reverse
            set_paths = []
            for path in reversed(paths):
                set_paths.append(imas_set(ids, path, ods[path], None, allocate=True))
            set_paths = filter(None, set_paths)

            # assign the data
            for path in set_paths:
                printd('writing %s' % l2i(path))
                imas_set(ids, path, ods[path], True)

            # actual write of IDS data to IMAS database
            for ds in ods.keys():
                if ds in add_datastructures.keys():
                    continue
                printd("ids.%s.put(0)" % ds, topic='imas_code')
                getattr(ids, ds).put(0)

        finally:
            # close connection to IMAS database
            printd("ids.close()", topic='imas_code')
            ids.close()

    return set_paths


@codeparams_xml_load
def load_omas_imas(user=os.environ.get('USER', 'dummy_user'), machine=None, pulse=None, run=0, paths=None,
                   imas_version=os.environ.get('IMAS_VERSION', omas_rcparams['default_imas_version']),
                   skip_uncertainties=False, skip_ggd=True, verbose=True):
    """
    Load OMAS data from IMAS

    NOTE: Either specify all or none of `user`, `machine`, `imas_version`
    If none of them are specified then use `imasdb` command to set the `MDSPLUS_TREE_BASE_?` environmental variables

    :param user: IMAS username

    :param machine: IMAS machine

    :param pulse: IMAS pulse

    :param run: IMAS run

    :param paths: list of paths to load from IMAS

    :param imas_version: IMAS version

    :param skip_uncertainties: do not load uncertain data

    :param skip_ggd: do not load ggd structure

    :param verbose: print loading progress

    :return: OMAS data set
    """

    if pulse is None or run is None:
        raise Exception('`pulse` and `run` must be specified')

    printd('Loading from IMAS (user:%s machine:%s pulse:%d run:%d, imas_version:%s)' % (user, machine, pulse, run, imas_version), topic='imas')

    try:
        ids = imas_open(user=user, machine=machine, pulse=pulse, run=run, new=False, imas_version=imas_version)

    except ImportError:
        if not omas_rcparams['allow_fake_imas_fallback']:
            raise
        filename = os.sep.join([omas_rcparams['fake_imas_dir'], '%s_%s_%d_%d_v%s.pkl' % (user, machine, pulse, run, imas_versions.get(imas_version, imas_version))])
        printe('Overloaded load_omas_imas: %s' % filename)
        from . import load_omas_pkl
        ods = load_omas_pkl(filename)

    else:

        try:
            # if paths is None then figure out what IDS are available and get ready to retrieve everything
            if paths is None:
                requested_paths = [[structure] for structure in list_structures(imas_version=imas_version)]
            else:
                requested_paths = list(map(p2l, paths))

            # fetch relevant IDSs and find available signals
            fetch_paths = []
            for ds in numpy.unique([p[0] for p in requested_paths]):
                if ds in add_datastructures.keys():
                    continue
                if not hasattr(ids, ds):
                    if verbose:
                        print('| ', ds)
                    continue
                # ids fetching
                printd("ids.%s.get()" % ds, topic='imas_code')
                try:
                    getattr(ids, ds).get()
                except ValueError as _excp:
                    print('x ', ds) # not sure why some IDSs fail on .get()... it's not about them being empty
                    continue
                if getattr(ids, ds).ids_properties.homogeneous_time != -999999999:
                    if verbose:
                        print('* ', ds)
                    fetch_paths += filled_paths_in_ids(ids, load_structure(ds, imas_version=imas_version)[1], [], [], requested_paths, skip_ggd=skip_ggd)
                else:
                    if verbose:
                        print('- ', ds)
            joined_fetch_paths = list(map(l2i, fetch_paths))

            # build omas data structure
            ods = ODS(imas_version=imas_version, consistency_check=False)
            for k, path in enumerate(fetch_paths):
                if path[-1].endswith('_error_upper') or path[-1].endswith('_error_lower') or path[-1].endswith('_error_index'):
                    continue
                if verbose and (k % 100 == 0 or k == len(fetch_paths) - 1):
                    print('Loading {0:3.3f}%'.format(100 * float(k) / (len(fetch_paths) - 1)))
                # get data from IDS
                data = imas_get(ids, path, None)
                # continue for empty data
                if data is None:
                    continue
                # add uncertainty
                if not skip_uncertainties and l2i(path[:-1] + [path[-1] + '_error_upper']) in joined_fetch_paths:
                    stdata = imas_get(ids, path[:-1] + [path[-1] + '_error_upper'], None)
                    if stdata is not None:
                        try:
                            data = uarray(data, stdata)
                        except uncertainties.core.NegativeStdDev as _excp:
                            printe('Error loading uncertainty for %s: %s' % (l2i(path), repr(_excp)))
                # assign data to ODS
                ods[path] = data

        finally:
            # close connection to IMAS database
            printd("ids.close()", topic='imas_code')
            ids.close()

    if paths is None:
        ods.setdefault('dataset_description.data_entry.user', unicode(user))
        ods.setdefault('dataset_description.data_entry.machine', unicode(machine))
        ods.setdefault('dataset_description.data_entry.pulse', int(pulse))
        ods.setdefault('dataset_description.data_entry.run', int(run))
        ods.setdefault('dataset_description.imas_version', unicode(imas_version))

    try:
        ods.consistency_check = True
    except LookupError as _excp:
        printe(repr(_excp))

    return ods


def browse_imas(user=os.environ.get('USER', 'dummy_user'), pretty=True, quiet=False,
                user_imasdbdir=os.sep.join([os.environ['HOME'], 'public', 'imasdb'])):
    '''
    Browse available IMAS data (machine/pulse/run) for given user

    :param user: user (of list of users) to browse. Browses all users if None.

    :param pretty: express size in MB and time in human readeable format

    :param quiet: print database to screen

    :param user_imasdbdir: directory where imasdb is located for current user (typically $HOME/public/imasdb/)

    :return: hierarchical dictionary with database of available IMAS data (machine/pulse/run) for given user
    '''
    # if no users are specified, find all users
    if user is None:
        user = glob.glob(user_imasdbdir.replace('/%s/' % os.environ['USER'], '/*/'))
        user = list(map(lambda x: x.split(os.sep)[-3], user))
    elif isinstance(user, basestring):
        user = [user]

    # build database for each user
    imasdb = {}
    for username in user:
        imasdb[username] = {}
        imasdbdir = user_imasdbdir.replace('/%s/' % os.environ['USER'], '/%s/' % username).strip()

        # find MDS+ datafiles
        files = list(recursive_glob('*datafile', imasdbdir))

        # extract machine/pulse/run from filename of MDS+ datafiles
        for file in files:
            tmp = file.split(os.sep)
            if not re.match('ids_[0-9]{5,}.datafile', tmp[-1]):
                continue
            pulse_run = tmp[-1].split('.')[0].split('_')[1]
            pulse = int(pulse_run[:-4])
            run = int(pulse_run[-4:])
            machine = tmp[-4]

            # size and data
            st = os.stat(file)
            size = st.st_size
            date = st.st_mtime
            if pretty:
                import time
                size = '%d Mb' % (int(size / 1024 / 1024))
                date = time.strftime('%d/%m/%y - %H:%M', time.localtime(date))

            # build database
            if machine not in imasdb[username]:
                imasdb[username][machine] = {}
            imasdb[username][machine][pulse, run] = {'size': size, 'date': date}

    # print if not quiet
    if not quiet:
        pprint(imasdb)

    # return database
    return imasdb


def load_omas_iter_scenario(pulse, run=0, paths=None,
                            imas_version=os.environ.get('IMAS_VERSION', omas_rcparams['default_imas_version']),
                            verbose=True):
    """
    Load OMAS data set from ITER IMAS scenario database

    :param pulse: IMAS pulse

    :param run: IMAS run

    :param paths: list of paths to load from IMAS

    :param imas_version: IMAS version

    :param verbose: print loading progress

    :return: OMAS data set
    """
    # set MDSPLUS_TREE_BASE_? environment variables as per
    # imasdb /work/imas/shared/iterdb/3 ; env | grep MDSPLUS_TREE_BASE
    try:
        bkp_imas_environment = {}
        for k in range(10):
            if 'MDSPLUS_TREE_BASE_%d' % k in os.environ:
                bkp_imas_environment['MDSPLUS_TREE_BASE_%d' % k] = os.environ['MDSPLUS_TREE_BASE_%d' % k]
            os.environ['MDSPLUS_TREE_BASE_%d' % k] = '/work/imas/shared/iterdb/3/%d' % k

        # load data from imas
        ods = load_omas_imas(user=None, machine=None, pulse=pulse, run=run, paths=paths, imas_version=imas_version, verbose=verbose)

    finally:
        # restore existing IMAS environment
        for k in range(10):
            del os.environ['MDSPLUS_TREE_BASE_%d' % k]
            os.environ.update(bkp_imas_environment)

    return ods


def filled_paths_in_ids(ids, ds, path=None, paths=None, requested_paths=None, assume_uniform_array_structures=False, skip_ggd=True):
    """
    Taverse an IDS and list leaf paths (with proper sizing for arrays of structures)

    :param ids: input ids

    :param ds: hierarchical data schema as returned for example by load_structure('equilibrium')[1]

    :param requested_paths: list of paths that are requested

    :param assume_uniform_array_structures: assume that the first structure in an array of structures has data in the same nodes locations of the later structures in the array

    :param skip_ggd: do not traverse ggd structures

    :return: returns list of paths in an IDS that are filled
    """
    if path is None:
        path = []

    if paths is None:
        paths = []

    if requested_paths is None:
        requested_paths = []

    # leaf
    if not len(ds):
        # append path if it has data
        if imas_empty(ids) is not None:
            paths.append(path)
        return paths

    # keys
    keys = list(ds.keys())
    if keys[0] == ':':
        keys = range(len(ids))
        if len(keys) and assume_uniform_array_structures:
            keys = [0]

    # kid must be part of this list
    if len(requested_paths):
        request_check = [p[0] for p in requested_paths]

    # traverse
    for kid in keys:

        # skip ggd structures
        if skip_ggd and kid in ['ggd', 'grids_ggd']:
            continue

        propagate_path = copy.copy(path)
        propagate_path.append(kid)

        # generate requested_paths one level deeper
        propagate_requested_paths = requested_paths
        if len(requested_paths):
            if kid in request_check or (isinstance(kid, int) and ':' in request_check):
                propagate_requested_paths = [p[1:] for p in requested_paths if len(p) > 1 and (kid == p[0] or p[0] == ':')]
            else:
                continue

        # recursive call
        if isinstance(kid, basestring):
            subtree_paths = filled_paths_in_ids(getattr(ids, kid), ds[kid], propagate_path, [], propagate_requested_paths, assume_uniform_array_structures, skip_ggd=skip_ggd)
        else:
            subtree_paths = filled_paths_in_ids(ids[kid], ds[':'], propagate_path, [], propagate_requested_paths, assume_uniform_array_structures, skip_ggd=skip_ggd)
        paths += subtree_paths

        # assume_uniform_array_structures
        if assume_uniform_array_structures and keys[0] == 0:
            zero_paths = subtree_paths
            for key in range(1, len(ids)):
                subtree_paths = copy.deepcopy(zero_paths)
                for p in subtree_paths:
                    p[len(path)] = key
                paths += subtree_paths

    return paths


def through_omas_imas(ods, method=['function', 'class_method'][1]):
    """
    Test save and load OMAS IMAS

    :param ods: ods

    :return: ods
    """
    user = os.environ['USER']
    machine = 'ITER'
    pulse = 1
    run = 0

    if method == 'function':
        paths = save_omas_imas(ods, user=user, machine=machine, pulse=pulse, run=run, new=True)
        ods1 = load_omas_imas(user=user, machine=machine, pulse=pulse, run=run, paths=paths)
    else:
        paths = ods.save('imas', user=user, machine=machine, pulse=pulse, run=run, new=True)
        ods1 = ODS().load('imas', user=user, machine=machine, pulse=pulse, run=run, paths=paths)
    return ods1



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
OMAS plot overlays
==================
This example loads sample DIII-D overlay data and generates an overlays plots
"""

from __future__ import print_function, division, unicode_literals

from matplotlib.pyplot import show, gca
from omas import *
import numpy

# load sample data from S3
ods = load_omas_s3('DIIIDoverlay_sample', user='omas_shared')

# plot equilibrium and wall
ods.plot_equilibrium_CX()
# plot overlays
ods.plot_overlay(  # Values can be True to simply turn on or dict to pass keywords to specific overlay functions
    gas_injection={'colors': ['b', 'g', 'c']},  # , 'which_gas': ['GASA', 'LOB1']},
    bolometer={'labelevery': 1, 'mask': numpy.append(numpy.ones(24, bool), (numpy.arange(24) % 2).astype(bool)), 'linestyle': '--'},
    thomson_scattering={'labelevery': 4, 'color': 'r'},
    debug_all_plots=True,
)
gca().legend(loc='center left', bbox_to_anchor=(1, 0.5))
gca().set_aspect('equal')

# display figure
show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Scaling IMAS performance
========================
This example shows a scaling performance study for manipulating OMAS data in hierarchical or tensor format.

The **hierarchical organization of the IMAS data** structure can in some situations hinder IMAS's ability to efficiently manipulate large data sets.
This contrasts to the **multidimensional arrays (ie. tensors)** approach that is commonly used in computer science for high-performance numerical calculations.

Based on this observation OMAS implements a transformation that casts the data that is contained in the IMAS hierarchical structure as a list of tensors, by taking advantage of the homogeneity of grid sizes that is commonly found across arrays of structures.
Such transformation and a summary of the scaling results are illustrated here for an hypothetical IDS that has data organized as a series of time-slices:

.. figure:: ../images/odx_concept.png
  :align: center
  :width: 100%
  :alt: OMAS implements a transformation that casts the data that is contained in the IMAS hierarchical structure as a list of tensors
  :target: /.._images/odx_concept.png

The favorable scaling that is observed when representing IMAS data in tensor form makes a strong case for adopting it.
Implementing the same system as part of the IMAS backend storage of data and in memory representation would likely greatly benefit IMAS performance in many real-world applications.

The new tensors representation would also greatly simplify the integration of IMAS with a broad range of tools and numerical libraries that are commonly used across many fields of science.

Finally, the addition of an extra dimension to the tensors could be used to efficiently store multiple realizations of signals from a distribution function of uncertain quantities.
Such feature would enable support of uncertainty quantification workflows and Bayesian integrated data analyses within IMAS.

Scaling study in detail
-----------------------

OMAS can seamlessly use either hierarchical or tensor representations as the backend for storing data both in memory and on file, and transform from one format to the other.
The mapping function is generic and can handle nested hierarchical list of structures (not only in time).
Also OMAS can automatically determine which data can be collected across the hierarchical structure, which cannot, and seamlessly handle both at the same time.

The following diagram summarizes the tests performed in this scaling study.
Benchmarks show that most operations stemming from the hierarchical representation of the data scale linearly with with the number of time-slices in the sample IDS (**red markers** in the diagram), whereas operations that make only use of the tensor representation show little to no dependency on the dataset size (**green markers** in the diagram).
As a result the tensors representation can be several orders of magnitude faster than a hierarchical organization, even for datasets of modest size.

.. figure:: ../images/odx_flow.png
  :align: center
  :width: 50%
  :alt: OMAS can seamlessly use either hierarchical or tensor representations as backed for storing data both in memory and on file, and transform from one format to the other
  :target: /.._images/odx_flow.png

Scaling plots and code used for the benchmark follow:

"""

import os
import time
import tempfile
from omas import *
import numpy
from pprint import pprint
from matplotlib import pyplot

ods = ODS()
ods.sample_equilibrium()

max_n = 100
max_samples = 11
stats_reps = 10
samples = numpy.unique(list(map(int, numpy.logspace(0, numpy.log10(max_n), max_samples)))).tolist()
max_samples = len(samples)

times = {}
for type in ['H', 'T']:  # hierarchical or tensor
    for action in ['R', 'W', 'M', 'A', 'S', 'B']:  # Read, Write, Mapping, Array access, Stripe access, Bulk access
        times[type + action] = []

for n in samples:
    print('%d/%d'%(n,samples[-1]))

    # keep adding time slices to the data structure
    for k in range(len(ods['equilibrium.time_slice']), n):
        ods.sample_equilibrium(time_index=k)

    # hierarchical write to HDF5
    filename = tempfile.gettempdir() + os.sep + 'tmp.h5'
    t0 = time.time()
    save_omas_h5(ods, filename)
    times['HW'].append(time.time() - t0)

    # hierarchical read from HDF5
    t0 = time.time()
    load_omas_h5(filename)
    times['HR'].append(time.time() - t0)

    # hierarchical access to individual array
    t0 = time.time()
    for k in range(stats_reps):
        for kk in range(n):
            ods['equilibrium.time_slice.%d.profiles_1d.psi' % kk]
    times['HA'].append((time.time() - t0) / n / float(stats_reps))

    # hierarchical slice across the data structure
    t0 = time.time()
    for kk in range(n):
        ods['equilibrium.time_slice.:.profiles_1d.psi'][:, 0]
    times['HS'].append((time.time() - t0) / n)

    # hierarchical bulk access to data
    t0 = time.time()
    for k in range(stats_reps):
        ods['equilibrium.time_slice.:.profiles_1d.psi']
    times['HB'].append((time.time() - t0) / float(stats_reps))

    # hierarchical mapping to tensor
    t0 = time.time()
    odx = ods_2_odx(ods)
    times['HM'].append(time.time() - t0)

    filename = tempfile.gettempdir() + os.sep + 'tmp.ds'

    # tensor write to HDF5
    t0 = time.time()
    save_omas_dx(odx, filename)
    times['TW'].append(time.time() - t0)

    # tensor read from HDF5
    t0 = time.time()
    odx = load_omas_dx(filename)
    times['TR'].append(time.time() - t0)

    # tensor mapping to hierarchical
    t0 = time.time()
    ods = odx_2_ods(odx)
    times['TM'].append(time.time() - t0)

    # tensor access to individual array
    t0 = time.time()
    for k in range(stats_reps):
        for kk in range(n):
            odx['equilibrium.time_slice.%d.profiles_1d.psi' % kk]
    times['TA'].append((time.time() - t0) / n / float(stats_reps))

    # tensor slice across the data structure
    t0 = time.time()
    for k in range(stats_reps):
        for kk in range(n):
            odx['equilibrium.time_slice.:.profiles_1d.psi'][:, 0]
    times['TS'].append((time.time() - t0) / n / float(stats_reps))

    # tensor bulk access to data
    t0 = time.time()
    for k in range(stats_reps):
        for kk in range(n):
            odx['equilibrium.time_slice.:.profiles_1d.psi']
    times['TB'].append((time.time() - t0) / n / float(stats_reps))

# print numbers to screen
print(samples)
pprint(times)

# plot read/write scaling
pyplot.figure()
for type in ['H', 'T']:  # hierarchical or tensor
    for action in ['R', 'W']:  # Read, Write
        pyplot.loglog(samples, times[type + action], label=type + action, lw=1.5, ls=['-', '--']['H' in type])
pyplot.xlabel('# of Equilibrium Time Slices')
pyplot.ylabel('Time [s]')
pyplot.legend(loc='upper left', frameon=False)
pyplot.title('Read/Write', y=0.85, x=0.7)

# plot mapping scaling
pyplot.figure()
for type in ['H', 'T']:  # hierarchical or tensor
    for action in ['M']:  # Mapping
        pyplot.loglog(samples, times[type + action], label=type + action, lw=1.5, ls=['-', '--']['H' in type])
pyplot.xlabel('# of Equilibrium Time Slices')
pyplot.ylabel('Time [s]')
pyplot.legend(loc='upper left', frameon=False)
pyplot.title('Mapping', y=0.85, x=0.7)

# plot access scaling
pyplot.figure()
for type in ['H', 'T']:  # hierarchical or tensor
    for action in ['A', 'S', 'B']:  # Array access, Stripe access, Bulk access
        pyplot.loglog(samples, times[type + action], label=type + action, lw=1.5, ls=['-', '--']['H' in type])
pyplot.xlabel('# of Equilibrium Time Slices')
pyplot.ylabel('Time [s]')
pyplot.legend(loc='upper left', frameon=False)
pyplot.title('Access', y=0.85, x=0.7)

pyplot.show()



'''save/load from UDA routines

-------
'''

from __future__ import print_function, division, unicode_literals

from .omas_utils import *
from .omas_core import ODS

try:
    _pyyda_import_excp = None
    import pyuda
except ImportError as _excp:
    _pyyda_import_excp = _excp


    # replace pyuda class by a simple exception throwing class
    class pyuda(object):
        """Import error UDA class"""

        def __init__(self, *args, **kwargs):
            raise _pyuda_import_excp


def load_omas_uda(server=None, port=None, pulse=None, run=0, paths=None,
                  imas_version=os.environ.get('IMAS_VERSION', omas_rcparams['default_imas_version']),
                  skip_uncertainties=False, skip_ggd=True, verbose=True):
    '''
    Load UDA data to OMAS

    :param server: UDA server

    :param port: UDA port

    :param pulse: UDA pulse

    :param run: UDA run

    :param paths: list of paths to load from IMAS

    :param imas_version: IMAS version

    :param skip_uncertainties: do not load uncertain data

    :param skip_ggd: do not load ggd structure

    :param verbose: print loading progress

    :return: OMAS data set
    '''

    if pulse is None or run is None:
        raise Exception('`pulse` and `run` must be specified')

    if server is not None:
        pyuda.Client.server = server
    elif not os.environ['UDA_HOST']:
        raise pyuda.UDAException('Must set UDA_HOST environmental variable')

    if port is not None:
        pyuda.Client.port = port
    elif not os.environ['UDA_PORT']:
        raise pyuda.UDAException('Must set UDA_PORT environmental variable')

    # set this to get pyuda metadata (maybe of interest for future use):
    # pyuda.Client.set_property(pyuda.Properties.PROP_META, True)

    client = pyuda.Client()

    # if paths is None then figure out what IDS are available and get ready to retrieve everything
    if paths is None:
        requested_paths = [[structure] for structure in list_structures(imas_version=imas_version)]
    else:
        requested_paths = map(p2l, paths)

    available_ds = []
    for ds in numpy.unique([p[0] for p in requested_paths]):
        if ds in add_datastructures.keys():
            continue

        if uda_get(client, [ds, 'ids_properties', 'homogeneous_time'], pulse, run) is None:
            if verbose:
                print('- ', ds)
            continue
        if verbose:
            print('* ', ds)
        available_ds.append(ds)

    ods = ODS(consistency_check=False)
    for k, ds in enumerate(available_ds):
        filled_paths_in_uda(ods, client, pulse, run, load_structure(ds, imas_version=imas_version)[1],
                            path=[], paths=[], requested_paths=requested_paths,
                            skip_uncertainties=skip_uncertainties, skip_ggd=skip_ggd,
                            perc=[float(k) / len(available_ds) * 100, float(k + 1) / len(available_ds) * 100, float(k) / len(available_ds) * 100])
    ods.consistency_check = True
    ods.prune()
    if verbose:
        print()
    return ods


def filled_paths_in_uda(ods, client, pulse, run, ds, path, paths, requested_paths, skip_uncertainties, skip_ggd, perc=[0., 100., 0.]):
    '''
    Recursively traverse ODS and populate it with data from UDA

    :param ods: ODS to be filled

    :param client: UDA client

    :param pulse: UDA pulse

    :param run: UDA run

    :param ds: hierarchical data schema as returned for example by load_structure('equilibrium')[1]

    :param path: []

    :param paths: []

    :param requested_paths: list of paths that are requested

    :param skip_uncertainties: do not load uncertain data

    :param skip_ggd: do not load ggd structure

    :return: filled ODS
    '''
    # leaf
    if not len(ds):
        return paths

    # keys
    keys = list(ds.keys())
    if keys[0] == ':':
        n = uda_get_shape(client, path, pulse, run)
        if n is None:
            return paths
        keys = range(n)

    # kid must be part of this list
    if len(requested_paths):
        request_check = [p[0] for p in requested_paths]

    # traverse
    n = float(len(keys))
    for k, kid in enumerate(keys):

        # skip ggd structures
        if skip_ggd and kid in ['ggd', 'grids_ggd']:
            continue

        if isinstance(kid, basestring):
            if skip_uncertainties and kid.endswith('_error_upper'):
                continue
            if kid.endswith('_error_lower') or kid.endswith('_error_index'):
                continue
            kkid = kid
        else:
            kkid = ':'

        # leaf
        if not len(ds[kkid]):
            # append path if it has data
            data = uda_get(client, path + [kid], pulse, run)
            if data is not None:
                # print(l2o(path))
                ods[kid] = data
                paths.append(path + [kid])

            pp = perc[0] + (k + 1) / n * (perc[1] - perc[0])
            if (pp - perc[2]) > 2:
                perc[2] = pp
                print('\rLoading: %3.1f%%' % pp, end='')

        propagate_path = copy.copy(path)
        propagate_path.append(kid)

        # generate requested_paths one level deeper
        propagate_requested_paths = requested_paths
        if len(requested_paths):
            if kid in request_check or (isinstance(kid, int) and ':' in request_check):
                propagate_requested_paths = [p[1:] for p in requested_paths if len(p) > 1 and (kid == p[0] or p[0] == ':')]
            else:
                continue

        # recursive call
        pp0 = perc[0] + k / n * (perc[1] - perc[0])
        pp1 = perc[0] + (k + 1) / n * (perc[1] - perc[0])
        pp2 = perc[2]
        paths = filled_paths_in_uda(ods[kid], client, pulse, run, ds[kkid], propagate_path, [], propagate_requested_paths, skip_uncertainties, skip_ggd, [pp0, pp1, pp2])

    # generate uncertain data
    if not skip_uncertainties and isinstance(ods.omas_data, dict):
        for kid in list(ods.omas_data.keys()):
            if kid.endswith('_error_upper') and kid[:-len('_error_upper')] in ods.omas_data:
                try:
                    if isinstance(ods[kid], ODS):
                        pass
                    elif isinstance(ods[kid], float):
                        ods[kid[:-len('_error_upper')]] = ufloat(ods[kid[:-len('_error_upper')]], ods[kid])
                    else:
                        ods[kid[:-len('_error_upper')]] = uarray(ods[kid[:-len('_error_upper')]], ods[kid])
                    del ods[kid]
                except Exception as _excp:
                    printe('Error loading uncertain data: %s' % kid)
    return paths


def uda_get_shape(client, path, pulse, run):
    '''
    Get the number of elements in a structure of arrays

    :param client: pyuda.Client object

    :param path: ODS path expressed as list

    :param pulse: UDA pulse

    :param run: UDA run

    :return: integer
    '''
    return uda_get(client, path + ['Shape_of'], pulse, run)


def offset(path, off):
    '''
    IMAS UDA indexing starts from one

    :param path: ODS path expressed as list

    :param off: offset to apply

    :return: path with applied offset
    '''
    return [p if isinstance(p, basestring) else p + off for p in path]


def uda_get(client, path, pulse, run):
    '''
    Get the data from UDA

    :param client: pyuda.Client object

    :param path: ODS path expressed as list

    :param pulse: UDA pulse

    :param run: UDA run

    :return: data
    '''
    try:
        location = l2o(offset(path, +1)).replace('.', '/')
        tmp = client.get(location, pulse)
        if isinstance(tmp, pyuda._string.String):
            return tmp.str
        else:
            return tmp.data
    except pyuda.UDAException:
        return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
UDA Universal Data Access
=========================
This example illustrates how OMAS can load machines
experimental data stored in IMAS format via UDA
"""

from omas import *
from pprint import pprint

# MAST UDA server (restricted access)
server = 'idam0.mast.ccfe.ac.uk'
port = 56563

pulse = 30420

ods = load_omas_uda(server=server,
                    port=port,
                    pulse=pulse,
                    paths=['magnetics'])

pprint(ods.pretty_paths())



try:
    from test_omas_core import *
    from test_omas_physics import *
    from test_omas_plot import *
    from test_omas_suite import *
    from test_omas_utils import *
except ImportError:
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Input data processing
=====================
OMAS supports user defined input data processing
"""

from omas import *


# user defined function that takes input string and converts it to python types
def robust_eval(string):
    import ast
    try:
        return ast.literal_eval(string)
    except:
        return string


# assign data and process it with user defined function
ods = ODS(consistency_check=False)
with omas_environment(ods, input_data_process_functions=[robust_eval]):
    ods['int'] = '1'
    ods['float'] = '1.0'
    ods['str'] = 'bla'
    ods['complex'] = '2+1j'

# test
for item in ods:
    print(ods[item], type(ods[item]))
    assert isinstance(ods[item], eval(item))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
OMAS handling of code parameters
==================

This example shows how XML code.parameters are handled withing OMAS

NOTE: OMAS will save code.parameters as XML only when saving to IMAS
All other save methods retain the code.parameters tree structure
"""

from pprint import pprint
from omas import *
import tempfile
import os

# Assign code parameters using dictionary approach
ods = ODS()
ods['equilibrium.code.parameters'] = CodeParameters()
ods['equilibrium.code.parameters']['test'] = {}
ods['equilibrium.code.parameters']['test']['parameter1'] = 1
ods['equilibrium.code.parameters']['test']['parameter2'] = 'hello'
with omas_environment(ods, xmlcodeparams=True):
    pprint(ods)

# Load code parameters from XML file
ods = ODS()
ods['equilibrium.time'] = [0.0]
ods['equilibrium.code.parameters'] = CodeParameters(imas_json_dir + '/../samples/input_gray.xml')

# In OMAS code.parameters are represented as dicts
pprint(ods)
# the XML representation can be set with an environment
with omas_environment(ods, xmlcodeparams=True):
    pprint(ods)
pprint(ods)

# The IMAS save is the only one that forces saving code.parameters as XML
# All other save methods maintain the dictionary structure
omas_rcparams['allow_fake_imas_fallback'] = True

# Save to IMAS
# code.parameters are saved as XML
print('=' * 20)
print(' Writing data to IMAS')
print('=' * 20)
save_omas_imas(ods, machine='ITER', pulse=1, new=True)

# Load from IMAS
# code.parameters are loaded as XML
print('=' * 20)
print(' Reading data from IMAS')
print('=' * 20)
ods1 = load_omas_imas(machine='ITER', pulse=1)

# code.parameters are dictionaries in the ODS
pprint(ods['equilibrium.code.parameters'])



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Sample OMAS data
================
OMAS provides a way to populate ODSs with sample data.
This is often useful for testing/debugging purposes.
"""

from __future__ import print_function, division, unicode_literals

import os

os.environ['OMAS_DEBUG_TOPIC'] = '*'

from omas import *

# list functions that fill ODS with sample data
ods_sample_methods = []
for item in dir(ODS):
    if item.startswith('sample_'):
        ods_sample_methods.append(item)
print("ods_sample_methods = %s" % ods_sample_methods)

ods = ODS()

# add sample equilibrium data to an ODS
ods.sample_equilibrium()

# alternatively, ods_sample() generates a sample ODS that is populated by all `ODS.sample_` methods
ods = ods_sample()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
OMFIT classes and plotting
==========================
How to generate ODSs from OMFITclasses, and use OMAS plot methods
This example only works with python 2.7 since OMFIT is only Python 2.7 compatible
"""

from __future__ import print_function, division, unicode_literals

from matplotlib.pyplot import show

from omfit.classes.omfit_eqdsk import OMFITgeqdsk, OMFITsrc
from omfit.classes.omfit_gacode import OMFITgacode
from omas import *

# =====================
# plotting equilibrium
# =====================
# read gEQDSK file in OMFIT
geq = OMFITgeqdsk(OMFITsrc + '/../samples/g133221.01000')
geq['fluxSurfaces'].load()

# convert gEQDSK to OMAS data structure
ods = geq.to_omas()

# omas plots using functional approach
omas_plot.equilibrium_summary(ods, linewidth=1)

# omas plots using object-oriented approach
# ods.plot_equilibrium_summary(linewidth=1)

# =====================
# plotting profiles
# =====================
# read input.profiles file in OMFIT
ip = OMFITgacode(OMFITsrc + '/../samples/input.profiles')

# convert input.profiles to OMAS data structure
ods = ip.to_omas()

# omas plots using functional approach
# omas_plot.core_profiles_summary(ods)

# omas plots using object-oriented approach
ods.plot_core_profiles_summary()

show()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Access data across multiple ODSs
================================
ODS with consistency_check=False can be used to traverse data across multiple ODSs
"""

from omas import *

# define a ods to collect multiple ODSs
master_ods = ODS(consistency_check=False)

# populate master_ods with a 5x3 grid of ODSs each having their data
for k in range(5):
    for j in range(3):
        ods = ODS(consistency_check=False)
        ods['data'] = k
        master_ods[k][j] = ods

# print data aggregating across ODSs
print(master_ods[':.:.data'])

###################################
# This will return::
#
#     [[0 0 0]
#      [1 1 1]
#      [2 2 2]
#      [3 3 3]
#      [4 4 4]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
showcase paths
==============
This example shows how OMAS supports dynamic path crection using different syntaxes.

.. figure:: ../images/dynamic_path_testimonial.png
  :align: center
  :width: 100%
  :alt: What people say about OMAS dynamic path creation
  :target: ../_images/dynamic_path_testimonial.png
"""

from __future__ import print_function, division, unicode_literals
import numpy

from omas import *

ods = ODS()

# without dynamic path creation one must use Python approach to create nested dictionaries
# this can be extremely tedious!
ods.dynamic_path_creation = False
ods['equilibrium'] = ODS()
ods['equilibrium']['time_slice'] = ODS()
ods['equilibrium']['time_slice'][0] = ODS()
ods['equilibrium']['time_slice'][0]['time'] = 1000
assert (ods['equilibrium']['time_slice'][0]['time'] == 1000.)

# Dynamic path creation (True by default) makes life easier.
# NOTE: OMAS supports different data access syntaxes
ods.dynamic_path_creation = True

# access data as dictionary
ods['equilibrium']['time_slice'][0]['time'] = 1000.
assert (ods['equilibrium']['time_slice'][0]['time'] == 1000.)

# access data as string
ods['equilibrium.time_slice.1.time'] = 2000.
assert (ods['equilibrium.time_slice.1.time'] == 2000.)

# access data as string (square brackets for arrays of structures)
ods['equilibrium.time_slice[2].time'] = 3000.
assert (ods['equilibrium.time_slice[2].time'] == 3000.)

# access data with path list
ods[['equilibrium', 'time_slice', 3, 'time']] = 4000.
assert (ods[['equilibrium', 'time_slice', 3, 'time']] == 4000.)

# access data with mix and match approach
ods['equilibrium']['time_slice.4.time'] = 5000.
assert (ods['equilibrium']['time_slice.4.time'] == 5000.)

# =============
# Data slicing
# =============

# classic ways to access data across an array of structures
data = []
for k in ods['equilibrium.time_slice'].keys():
    data.append(ods['equilibrium.time_slice'][k]['time'])
data = numpy.array(data)
assert (numpy.all(data == numpy.array([1000., 2000., 3000., 4000., 5000.])))

# access data across an array of structures via data slicing
data = ods['equilibrium.time_slice.:.time']
assert (numpy.all(data == numpy.array([1000., 2000., 3000., 4000., 5000.])))

# =========================
# .setdefault() and .get()
# =========================

# like for Python dictionaries .setdefault() will set an entry with its
# default value (second argument) only if that entry does not exists already
ods = ODS()
ods['equilibrium.time_slice.0.global_quantities.ip'] = 6
ods.setdefault('equilibrium.time_slice.0.global_quantities.ip', 5)
assert (ods['equilibrium.time_slice.0.global_quantities.ip'] == 6)
ods = ODS()
ods.setdefault('equilibrium.time_slice.0.global_quantities.ip', 5)
assert (ods['equilibrium.time_slice.0.global_quantities.ip'] == 5)

# like for Python dictionaries .get() return the value of an entry or its
# default value (second argument) if that does not exists
ods = ODS()
ods.get('equilibrium.time_slice.0.global_quantities.ip', 5)
assert ('equilibrium.time_slice.0.global_quantities.ip' not in ods)

# ========
# Cleanup
# ========

# Dynamic path creation can leave empty trees behind in case of bad IMAS location is entered.
# This is inevitable when splitting the path in individual pieces.
# These leftover are often innocuous, and occur only during the development stages.
# Use of .prune() to clean empty branches left behind from dynamic path creation
ods = ODS()
try:
    ods['equilibrium']['time_slice'][0]['global_quantities']['not_valid'] = 6
except LookupError:
    n = ods.prune()
    assert (n == 4)
    assert (len(ods) == 0)

# Note that single string access does not leave empty branches
ods = ODS()
try:
    ods['equilibrium.time_slice.0.global_quantities.asdasd'] = 6
except LookupError:
    assert (len(ods) == 0)
